/************************************************************************\
** PI3CONV 1.0 - "C" language source file for MEGAMAX C compiler.	**
**									**
** Operating system:	GEM/GEMDOS (Uses VDI & XBIOS).			**
** Computer:		Atari 520ST & 1040ST, Color or Monochrome.	**
**									**
** Description: Converts High-Resolution DEGAS-format ".PI3" files to 	**
**		NEOCHROME ".NEO" format.				**
**									**
** Notes:	High-Resolution halftones converted to low-resolution	**
**		greyscale.  Picture is virtually identical to Monochrome**
**		display when viewed from 6 feet away, and presentable	**
**		at close-range.  Uses 7 shades of grey.			**
**									**
** Assembly procedure:							**
**									**
**	mmcc pi3conv.c							**
**	mmlink pi3conv.o -o pi3conv.prg					**
**									**
**	(A WHOLE lot easier than ALCYON C, eh?)				**
**									**
** Use: (1) double-click (open) the PI3CONV.PRG file.			**
**	(2) Click on my name in the title alert (or press return).	**
**	(3) Use the item-selector to select .PI3 files for conversion.	**
**	(4) Hit the "CANCEL" button in the item-selector to return to	**
**	    Desktop.							**
**									**
** Conversion times:							**
**	about 17 seconds, floppy to floppy,				**
**	about 4 seconds+1/per megabyte, hard disk to hard disk,		**
**	about 3 seconds RAM disk to RAM disk.				**
**									**
** Compile time (To executable program):				**
**	about 15 seconds, RAM disk to RAM disk, compiler/linker on	**
**			  hard disk.  (AMAZING!)			**
**									**
\************************************************************************/

#include <osbind.h>
#include <define.h>

#define TITLE_ALERT "[0][       PI3CONV 1.0|Converts High-Res. DEGAS\277|\
(.PI3) files to .NEO files.| |(Public Domain: 04/04/86)]\
[ By: Ken Settle ]"	/* title alert text */

char pathbuf[80]="";
char fname[80]="";
char newstuff[92]="        .   "; /* required by NEO, I don't know why. */

/* halftone to greyscale conversion array: */

char convert[]={1,2,3,4,2,3,4,5,3,4,5,6,4,5,6,7};

/* default pallete for converted .NEO output files: */

int newpallete[18]={0,0,0,0,0x222,0x333,0x444,0x555,0x666,
			0x777,0,0,0,0,0,0,0,0};

int invert;			/* was it a inverted picture???? */

int gl_apid;			/* process ID # */
int oldrez;
long oldphys,oldvbl;

int color[16];			/* array for storage of original pallette */

long error;
int handle;
int handleo;
long length;

char *buffer;			/* display buffer */

int phys_handle,ret,vhandle=0;

int contrl[12];		/* GEM binding arrays */
int intin[128];
int ptsin[128];
int intout[128];
int ptsout[128];

int work_in[11];	/* Input to GSX parameter array */
int work_out[57];	/* Output from GSX parameter array */

#define SP A7
#define dbra dbf

hidemouse()		/* hide mouse, open VDI workstation if necessary */
{register int index;

 if(!vhandle)
 {
  for(index=0;index<10;work_in[index++]=1);
  work_in[10]=2;			/* use raster coordinates */
  vhandle=phys_handle;
  v_opnvwk(work_in,&vhandle,work_out);

  v_hide_c(vhandle);			/* hide cursor */
  v_clsvwk(vhandle);
  vhandle=0;				/* if no virtual workspace,handle=0 */
 }
 else
  v_hide_c(vhandle);
}
 
showmouse()		/* show mouse, open VDI workstation if necessary */
{register int index;

 if(!vhandle)				/* if no VDI workstation open */
 {
  for(index=0;index<10;work_in[index++]=1); /* open one */
  work_in[10]=2;			/* use raster coordinates */
  vhandle=phys_handle;
  v_opnvwk(work_in,&vhandle,work_out);

  v_show_c(vhandle,0);			/* show cursor (regardless of # hides)*/
  v_clsvwk(vhandle);			/* close workstation again */
  vhandle=0;				/* if no virtual workspace,handle=0 */
 }
 else
  v_show_c(vhandle,0);			/* show cursor (regardless of # hides)*/
}

conv_pi3(source,dest)			/* convert .PI3 file to easier format */
char *source,*dest;
{asm
 {
	movem.l	D4-D7,-(SP)
	move.l	source(A6),A0		;A0.L points to screen base
	move.l	dest(A6),A1		;A1.L points to bottom right
	move.w	#200-1,D5		;D5.W=# lines we get
loop3:
	move.w	#40-1,D7		;D7.W=# times we get 8 pixels (-1/DBRA)
loop1:
	moveq	#8-1,D6			;D6.W=# pixels to process (-1/DBRA)
	move.w	(A0)+,D0		;D0.W=plane 0 & 1
	move.w	78(A0),D1		;D2.W=plane 2 & 3
loop2:
	moveq	#0,D4			;D4.W=pixel color accumulator

	roxl.w	#1,D1			;get plane 3
	roxl.w	#1,D4			;shift into pixel
	roxl.w	#1,D1			;get plane 2
	roxl.w	#1,D4			;shift into pixel
	roxl.w	#1,D0			;get plane 1
	roxl.w	#1,D4			;shift into pixel
	roxl.w	#1,D0			;get plane 0
	roxl.w	#1,D4			;shift into pixel

	move.b	D4,(A1)+		;store pixel byte
	dbra	D6,loop2

	dbra	D7,loop1

	lea.l	80(A0),A0		; skip to next line
	dbra	D5,loop3

	movem.l	(SP)+,D4-D7
 }
}

conv_back(source,dest)			/* convert back to ST Graphics format */
char *source,*dest;
{asm
 {
	movem.l	D4-D7,-(SP)
	move.l	source(A6),A0		;A0.L points to screen base
	move.l	dest(A6),A1		;A1.L points to bottom right
	move.w	#4000-1,D7		;D7.W=# times we get 16 pixels (-1/DBRA)
loop1:
	moveq	#15,D6			;D6.W=# pixels to process (-1/DBRA)
loop2:
	move.b	(A0)+,D4		;D4.B=pixel color
	roxr.w	#1,D4			;get plane 0
	roxl.w	#1,D0			;shift into output data
	roxr.w	#1,D4			;get plane 1
	roxl.w	#1,D1			;shift into output data
	roxr.w	#1,D4			;get plane 2
	roxl.w	#1,D2			;shift into output data
	roxr.w	#1,D4			;get plane 3
	roxl.w	#1,D3			;shift into output data
	dbra	D6,loop2

	movem.w	D0-D3,(A1)		;store pixel data
	addq.l	#8,A1

	dbra	D7,loop1

	movem.l	(SP)+,D4-D7
 }
}

read_the_file(fname)			/* read the .PI3 file into buffer */
char *fname;
{register int indexr;

 hidemouse();

 if(oldrez != 2)			/* if not on a high-res display */
  Setscreen(-1L,-1L,0);			/* force low resolution */

 error=Fopen(fname,0);			/* open picture file to read */
 if(error < 0)
  return(FALSE);

 handle=(int)error;
 length=Fread(handle,34L,buffer); 	/* read pallete colors into buffer */
 if(length < 0) 
  return(FALSE);

 invert=buffer[3];			/* true if we should invert */
 
 if(invert)				/* invert pallette to match hires */
  for(indexr=3; indexr < 18; indexr++)
  newpallete[indexr] ^= 0x777;

 if(oldrez != 2)			/* if not hires */
  Setpallete(&newpallete[2]);		/* set pallette from new data */
 else
  Setpallete(&buffer[2]);		/* else-set pallette from .PI3 file */

 length=Fread(handle,32000L,oldphys);	/* read rest of file to screen */
 if(length < 31999)
  return(FALSE);

 error=Fclose(handle);			/* close input file */
 if(error < 0)
  return(FALSE);
 else
  return(TRUE);
}

write_the_file(fname)			/* write display to .NEO file */
char *fname;
{register int indexr;

 hidemouse();
 
 for(indexr=12; indexr < 92; indexr++)	/* fill out newstuff with zeros */
  newstuff[indexr] = '\0';

 indexr=strlen(fname);			/* get length of last filename */

 while(indexr && (fname[indexr-1] != '.'))
  indexr--;				/* find last "." in filename */

 fname[indexr] = '\0';			/* wipe out extension from filename */
 strcat(fname,"NEO");			/* add .NEO extension to filename */

 error=Fcreate(fname,0);		/* open picture file to write */
 if(error < 0)
  return(FALSE);

 handle=(int)error;
 length=Fwrite(handle,36L,newpallete);	/* write pallette colors */
 if(length < 36) 
  return(FALSE);

 length=Fwrite(handle,92L,newstuff);	/* write extra stuff for .NEO */
 if(length < 92)
  return(FALSE);

 length=Fwrite(handle,32000L,oldphys);		/* write display */
 if(length < 32000) 
  return(FALSE);

 error=Fclose(handle);				/* close output file */
 if(error < 0)
  return(FALSE);
 else
  return(TRUE);
}

/* these 2 routines from DEMO.C (DRI) & modified for GEMDOS */

get_path(path,spec)	/* get directory path name for file selector */
register char *path,*spec;
{
	path[0]=(Dgetdrv())+'A';	/* get the default drive */
	path[1]=':';			/* drive seperator */
	Dgetpath(&path[2],0);		/* get the rest of default pathlist */
	if(strlen(path) > 2)
		strcat(path,"\\");	/* add "\" if inside a folder */
	else
		path[2]='\0';		/* else end string after drive name */

	strcat(path,spec);		/* add file selector wildcard */
}

add_file_name(pathname,filename)	/* add filename to end of pathname */
register char *pathname,*filename;
{
 register char c;
 register int  ii;

	ii = strlen(pathname);	/* ii points to end of pathname */

	while(ii && (((c = pathname[ii-1]) != '\\') && (c != ':')))
		ii--;		/* find last "\" or ":" in pathlist */

	pathname[ii] = '\0';	/* wipe out wildcard/extension from pathlist */
	strcat(pathname,filename);	/* add filename to end of pathlist */
}

error_exit()			/* display file read error (alert!) */
{register int i;

 for(i=0;i < 16;i++)	/* restore all color REGs */
  Setcolor(i,color[i]);

 showmouse();
 Setscreen(oldphys,oldphys,oldrez);/* restore original resolution (low/med) */
	form_alert(1,			/* display an error alert */
  "[3][Error!|Picture file is bad|or doesn't exist.][ Cancel ]");

	return(dofile());		/* retry the file selector (recursion)*/
}

dofile()		/* use item-selector to get .PI3 file & load it */
{int i,j;

	get_path(pathbuf,"*.PI3");	/* get the pathlist for file selector */

	j = fsel_input(pathbuf,fname,&i);	/* get file in default path */

	add_file_name(pathbuf,fname);		/* add filename to pathlist */

	if (i)				/* if not <CANCEL> */
	{  if (j)			/* if no error returned/fsel_input */
	   {  j = read_the_file(pathbuf); /* load file */
	      if (!j)			/* if file_load was OK */
	      	 return(error_exit());  /* and return with error alert */
	   }
	   else /* error in fsel_input */
	      return(error_exit());	/* exit with error */

	   return(TRUE);		/* we had no errors, file loaded. */
	}
	else /* they hit <CANCEL> in fsel_input */
	   return(FALSE);		/* exit with no errors <CANCEL> */
}

main()
{register int index;

 for(index=0;index < 16;index++)	/* get all color REGs */
  color[index]=Setcolor(index,-1);

 oldrez=Getrez();			/* get the current resolution */
 oldphys=Physbase();			/* and screen address */

 buffer = (char *)Malloc(65535L); 	/* get memory for conversion buffer */

 if(buffer)				/* if no memory allocation errors */
 {
  appl_init();				/* tell GEM we're here */

  phys_handle=graf_handle(&ret,&ret,&ret,&ret);	/* get handle to VDI wkspace */

  form_alert(1,TITLE_ALERT);		/* show user who we are */

  while(dofile())			/* while no errors & no cancel */
   pi3conv();				/* convert screen files */

  appl_exit();				/* done with GEM */
 }
}	/* done */

pi3conv()			/* convert .PI3 file (already loaded) */
{register int i;
 register long index;
 register long y;
 register char *buff=buffer;	/* make a fast local pointer to buffer */
 
 conv_pi3(oldphys,buff);	/* convert picture from PI3 format */

 for (y=0;y<64000;y++)		/* with every pixel on the display */
  buff[y]=convert[buff[y]];	/* convert halftones to grey-scale colors */

 conv_back(buffer,oldphys);	/* convert back to "displayable" .NEO format */

 index=write_the_file(pathbuf);	/* write the .NEO file */

 if(invert)			/* if pallete inverted, return it to normal */
  for(i=3; i < 18; i++)			/* kludge, kludge */
  newpallete[i] ^= 0xfff;

 for(i=0;i < 16;i++)		/* restore all color REGs */
  Setcolor(i,color[i]);

 Setscreen(oldphys,oldphys,oldrez);/* restore original resolution (low/med) */

 showmouse();

 if(!index)			/* if error writing output file, alert */
  form_alert(1,"[3][Error!|Output picture file is bad.][ Cancel ]");

}

EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE