**************************************************************************
*
* comprasm.s     -    Routines for the FAT compare util.
*
* 03/14/87 v1.0  -    Created original version, consisting of 
*                     routines to intercept hdv_rw and deal with
*                     forced media change on non-floppy devices, so that
*                     GEMDOS will throw away buffered directory sectors.
*                     (Routines extracted from zerodasm.s)
*
**************************************************************************

* Routines to deal with forced mediach on non-floppy devs.

          .globl    _install_mediach,_deinstall_mediach

buffer      = 6                         * Stack offsets to
count       = 10                        *  Rwabs()
device      = 14                        *  parameters.


oldvec:   dc.l      0                   * Old hdv_rw vector contents.
statbits  dc.l      0                   * Bit map of SAFE/UNSAFE status.

handler:  move.w    device(sp),d1       * Get device...
          cmp.w     #2,d1               * If device is floppy,
          blt       punt                * get outta here.
          ext.l     d1                  * Clean up d1 for bit ops.
          move.l    statbits,d2         * Get bitmap in reg for bit ops.
          tst.l     buffer(sp)          * Check attempt to force status,
          beq       forceit             * buffer addr = 0, go force it.
          btst.l    d1,d2               * Else, check current status,
          beq       punt                * If not forced, get out. 
          move.l    #-14,d0             * If forced, return changed 
reset:    bclr.l    d1,d2               * status, clear for next call,
          bra       return              * and return.
forceit:  move.l    #0,d0               * Forced change always works.
          tst.w     count(sp)           * If forced to SAFE, go reset
          beq       reset               * and return.
          bset.l    d1,d2               * If forced to UNSURE or UNSAFE
return:   move.l    d2,statbits         * make it UNSAFE, put it back
          rts                           * into memory from reg, return.
punt:     move.l    oldvec,a0           * To punt out of here, just jump
          jmp       (a0)                * through old vector.

_install_mediach: 
          clr.l     -(sp)               * Get into supervisor mode.
          move.w    #$20,-(sp)
          trap      #1
          move.l    $476,oldvec         * Install hook.
          move.l    #handler,$476
          move.l    d0,2(sp)            * Restore ssp.
          trap      #1                  * Back into user mode.
          addq.l    #6,sp
          rts

_deinstall_mediach:
          clr.l     -(sp)               * Get into supervisor mode.
          move.w    #$20,-(sp)
          trap      #1
          move.l    oldvec,$476         * De-Install hook.
          move.l    d0,2(sp)            * Restore ssp.
          trap      #1                  * Back into user mode.
          addq.l    #6,sp
          rts
          .end
