**************************************************************************
*
* zerodasm.S     -    Routines for the image backup util.
*
* 02/26/87 v1.0  -    Created original version.
* 03/12/87 v1.1  -    Added routines to intercept hdv_rw and deal with
*                     forced media change on non-floppy devices, so that
*                     GEMDOS will throw away buffered directory sectors.
*
*
**************************************************************************

*                                       * OFFSETS TO STACK PARAMETERS:
p_fatbuf   =        8                   *   Pointer to buffered fat
p_bpb      =        12                  *   Pointer to device bpb
*                                       * OFFSETS INTO BPB:
bpb_clsiz  =        2                   *   Number of sectors/cluster
bpb_datrec =        12                  *   Sector # of 1st data sector.
bpb_numcl  =        14                  *   # of clusters mapped by fat.
bpb_bflags =        17                  *   12/16 bit fat entries flag.

          .globl    _fixfat

_fixfat:  link      a6,#0
          movem.l   d3/a3-a4,-(sp)      * Save working registers.
          move.l    p_fatbuf(a6),a1     * a1 -> buffered FAT.
          move.l    p_bpb(a6),a2        * a2 -> device BPB.
          move.w    #2,d3               * Start at cl 2 (0 & 1 are rsvd).
          move.l    #put16bit,a4
          move.l    #get16bit,a3        * Make a3 into a pointer to a
          btst.b    #0,bpb_bflags(a2)   * routine to process the right
          bne       fixl1               * flavor of FAT entries (12 or
          move.l    #get12bit,a3        * 16 bit), for indirect jsr.
          move.l    #put12bit,a4
          
fixl1:    move.w    d3,d1               * Major loop point for FAT entries
          jsr       (a3)                * Go get entry, test return
          tst.l     d0                  * value.  A value of zero
          bmi       fixl2               * or neg indicates non-data
          beq       fixl2               * cluster go see if this is
          clr.w     d0
          jsr       (a4)
fixl2:    addq.w    #1,d3
          cmp.w     bpb_numcl(a2),d3
          ble       fixl1
          movem.l   (sp)+,d3/a3-a4
          unlk      a6
          rts

*------------------------------------------------------------------------

get16bit: clr.l     d0
          move.b    1(a1,d1.w),d0       * 16 bit is easy...get the word
          lsl.w     #8,d0               * indexed by clnum, and byte
          move.b    0(a1,d1.w),d0       * swap it into 68k format.  
          cmp.l     #$0000fff0,d0       * Check for bad sector, which
          blt       getret              * must be distinguished from the
          btst.l    #3,d0               * EOF marker due to stupidity at
          beq       getneg              * IBM (of all things!).  If it's
          bra       getret              * bad sector, go return a neg.

get12bit: clr.l     d0
          move.w    d1,d2               * To get a 12 bit entry, mult
          lsr.w     #1,d2               * the index by 1.5 (add half of
          add.w     d1,d2               * it to itself), and get the
          move.b    1(a1,d2.w),d0       * resulting (unaligned) word,
          lsl.w     #8,d0               * and byte swap it to 68k format.
          move.b    0(a1,d2.w),d0       * If the index is odd, it's we
          btst.l    #0,d1               * have to shift the swapped 
          beq       get12skp            * entry right by 4 bits, if it
          lsr.w     #4,d0               * is even, we just mask off the
get12skp: and.w     #$0fff,d0           * crufty bits, and leave it at
          cmp.w     #$0ff0,d0           * that.  After doctoring the 12
          blt       getret              * bit entry to a word, check to
          btst.l    #3,d0               * see if it is a bad sector,
          bne       getret              * so we can return a long 
getneg    or.l      #$ffff0000,d0       * negative if it is.

getret:   rts

*------------------------------------------------------------------------

put16bit: move.b    d0,0(a1,d1.w)
          lsr.w     #8,d0
          move.b    d0,0(a1,d1.w)
          rts

put12bit: movem.l   d3-d4,-(sp)
          move.w    d0,d3
          and.w     #$0fff,d3
          move.w    #$f000,d4
          btst.l    #0,d1
          beq       put12skp
          lsl.w     #4,d3
          move.w    #$000f,d4
put12skp: move.w    d1,d2               * To get a 12 bit entry, mult
          lsr.w     #1,d2               * the index by 1.5 (add half of
          add.w     d1,d2               * it to itself), and get the
          move.b    1(a1,d2.w),d0       * resulting (unaligned) word,
          lsl.w     #8,d0               * and byte swap it to 68k format.
          move.b    0(a1,d2.w),d0       * If the index is odd, it's we
          and.w     d4,d0
          or.w      d3,d0
          move.b    d0,0(a1,d2.w)
          lsr.w     #8,d0
          move.b    d0,1(a1,d2.w)
          movem.l   (sp)+,d3-d4
          rts

*------------------------------------------------------------------------

* Routines to deal with forced mediach on non-floppy devs.

          .globl    _install_hdv_rw,_deinstall_hdv_rw

buffer      = 6                         * Stack offsets to
count       = 10                        *  Rwabs()
device      = 14                        *  parameters.


oldvec:   dc.l      0                   * Old hdv_rw vector contents.
statbits  dc.l      0                   * Bit map of SAFE/UNSAFE status.

handler:  move.w    device(sp),d1       * Get device...
          cmp.w     #2,d1               * If device is floppy,
          blt       punt                * get outta here.
          ext.l     d1                  * Clean up d1 for bit ops.
          move.l    statbits,d2         * Get bitmap in reg for bit ops.
          tst.l     buffer(sp)          * Check attempt to force status,
          beq       forceit             * buffer addr = 0, go force it.
          btst.l    d1,d2               * Else, check current status,
          beq       punt                * If not forced, get out. 
          move.l    #-14,d0             * If forced, return changed 
reset:    bclr.l    d1,d2               * status, clear for next call,
          bra       return              * and return.
forceit:  move.l    #0,d0               * Forced change always works.
          tst.w     count(sp)           * If forced to SAFE, go reset
          beq       reset               * and return.
          bset.l    d1,d2               * If forced to UNSURE or UNSAFE
return:   move.l    d2,statbits         * make it UNSAFE, put it back
          rts                           * into memory from reg, return.
punt:     move.l    oldvec,a0           * To punt out of here, just jump
          jmp       (a0)                * through old vector.

_install_hdv_rw: 
          clr.l     -(sp)               * Get into supervisor mode.
          move.w    #$20,-(sp)
          trap      #1
          move.l    $476,oldvec         * Install hook.
          move.l    #handler,$476
          move.l    d0,2(sp)            * Restore ssp.
          trap      #1                  * Back into user mode.
          addq.l    #6,sp
          rts

_deinstall_hdv_rw:
          clr.l     -(sp)               * Get into supervisor mode.
          move.w    #$20,-(sp)
          trap      #1
          move.l    oldvec,$476         * De-Install hook.
          move.l    d0,2(sp)            * Restore ssp.
          trap      #1                  * Back into user mode.
          addq.l    #6,sp
          rts
          .end
