/**************************************************************************
 *
 * zerodir.c      -    Clear all files off of a volume by zero-ing the
 *                     FAT and root directory.  Preserve volume labels
 *                     and give the user the option to change/add labels.
 *                     If there are bad sectors flagged in the FAT, this
 *                     util will not zero them, they'll still be flagged.
 *
 * Author         -    Ian Lepore
 *
 * 02/26/87 v1.0  -    Created original version.
 * 03/12/87 v1.1  -    Added routines to intercept hdv_rw and deal with
 *                     forced media change on non-floppy devices, so that
 *                     GEMDOS will throw away buffered directory sectors.
 *
 **************************************************************************/

#include <osbind.h>
#include <gemdefs.h>
#include <obdefs.h>
#include "ZERODIR.H"

#define   say(x)    Cconws((x))
#define   Getbpb(a) bios(7,(a))

struct     {
           char fs_filler[21];
           char fs_attr;
           int  fs_time;
           int  fs_date;
           long fs_size;
           char fs_name[13];
           } fs_dta, *save_dta;

struct BPB {
           int bpb_recsiz,  bpb_clsiz,  bpb_clsizb,
               bpb_rdlen,   bpb_fsiz,   bpb_fatrec,
               bpb_datrec,  bpb_numcl,  bpb_bflags;
           } *p_bpb;

int  vdi_handle, contrl[12], 
     intin[128], ptsin[128],
     intout[128],ptsout[128],
     work_in[12] = {1,1,1,1,1,1,1,1,1,1,0,0},
     work_out[57];

char *p_buffer,
     workstr[5],
     drv_char[] = "DRIVE ?",
     workchr,
     vol_label[13] = "filename.ext",
     vol_path[4]   = "?:\\",
     wrk_path[18],
     dmysearch[] = "?:\\*.*",
     search[] = "?:\\*.*";

int  unsure,
     counter,
     drv_code,
     fst_sector,
     num_sectors;

long *p_work,
     drv_bits,
     maintree,
     warntree,
     infotree,
     stat;

char no_rsrc_alert[]    = "[3][ | |    Resource file not found!   | | ][ CANCEL ]",
     err_getbpb_alert[] = "[3][ Error obtaining device | parameters.  | | Volume remains unchanged. | ][ CANCEL ]",
     err_malloc_alert[] = "[3][ Error obtaining memory. | | Volume remains unchanged. | ][ CANCEL ]",
     err_ioerr_alert[]  = "[3][ I/O error occurred.     | | Disk may now need to be | formatted. ][ CANCEL ]",
     err_label_alert[]  = "[2][ | Volume is cleared, but error | occurred while writing label. | ][  OK  ]",
     ok_alldone_alert[] = "[0][ | | Completed successfully | | ][  OK  ]";
                                   
strcpy(d,s)
char *d,*s;
{
while (*d++ = *s++);
}

/***********************************************************************
 *
 * GEM initialization.
 *   This routine invokes installation of a special media change handler
 *   to deal with GEMDOS buffers.  It does typical AES/VDI init stuff.
 *   It determines which drives (other than A & B) live in the system,
 *   and enables the dialog box buttons for those drives.
 *
 ***********************************************************************/

prg_init()
{
int  counter;
long *ptr_tedinf;

install_hdv_rw();                       /* Install the mediach handler */

appl_init();
v_opnvwk(work_in,&vdi_handle,work_out);

if (!rsrc_load("ZERODIR.RSC"))              /* Load in resource file  */
     {
     form_alert(1,no_rsrc_alert);
     prg_exit();
     }

rsrc_gaddr(0,MAINTREE,&maintree);           /* Get addr of dial parents*/
rsrc_gaddr(0,WARNTREE,&warntree);
rsrc_gaddr(0,INFOTREE,&infotree);

ptr_tedinf = (*((long *)(warntree + ( WARNTEVL * sizeof(OBJECT)) + 12)));
*ptr_tedinf = vol_label;        /* Point text edit field to our string */
ptr_tedinf = (*((long *)(warntree + ( WARNTFDR * sizeof(OBJECT)) + 12)));
*ptr_tedinf = drv_char;        /* Point frmtd text field to our string */

drv_bits = Drvmap() >> 2;

counter = 0;
while (drv_bits)
     {
     if (drv_bits & 0x00000001)
          objst_reset(maintree, MAINBXC + counter++, DISABLED); 
     drv_bits >>= 1;
     }

graf_mouse(ARROW,0L);                       /* Set mouse to arrow.    */
}

/***********************************************************************
 *
 * Program cleanup and exit.
 *   This routine does typical AES/VDI cleanup, and de-installs the
 *   media change handler.
 *
 ***********************************************************************/

prg_exit()
{
v_clsvwk(vdi_handle);                        /* Close workstation.      */
appl_exit();                                 /* Say Goodbye to AES.     */
deinstall_hdv_rw();                          /* Get rid of mediach hndlr*/
Pterm(0);                                    /* And get outta here.     */

}


/************************************************************************
 *
 * Object manipulations. (Get, Set, or Reset bit in state flag of obj).
 *   Mostly stolen from the DRDOODLE source code.
 *
 ************************************************************************/

objst_set(tree, objindx, bit)
long tree;
int  objindx, bit;
{
int  cur_state;
cur_state = objst_get(tree, objindx);
( *((int *)(tree + objindx * sizeof(OBJECT) + 10)) = cur_state | bit);
}

objst_reset(tree, objindx, bit)
long tree;
int  objindx, bit;
{
int  cur_state;
cur_state = objst_get(tree, objindx);
( *((int *)(tree + objindx * sizeof(OBJECT) + 10)) = cur_state & ~bit);
}

objst_get(tree, objindx)
long tree;
int  objindx;
{
return( (int) *((int *)(tree + objindx * sizeof(OBJECT) + 10)) );
}

objfl_set(tree, objindx, bit)
long tree;
int  objindx, bit;
{
int  cur_flag;
cur_flag = objfl_get(tree, objindx);
( *((int *)(tree + objindx * sizeof(OBJECT) + 8)) = cur_flag | bit);
}

objfl_reset(tree, objindx, bit)
long tree;
int  objindx, bit;
{
int  cur_flag;
cur_flag = objfl_get(tree, objindx);
( *((int *)(tree + objindx * sizeof(OBJECT) + 8)) = cur_flag & ~bit);
}

objfl_get(tree, objindx)
long tree;
int  objindx;
{
return( (int) *((int *)(tree + objindx * sizeof(OBJECT) + 8)) );
}

/************************************************************************
 *
 * Handle a dialog, start to finish.  Return the index of the exit obj.
 *   More DRDOODLE plagarism.
 *
 ************************************************************************/

dial_do(cur_tree,defindx)
long cur_tree;
int  defindx;
{
int         xdial, ydial, wdial, hdial, exitobj;
static int  xgrowpt = 0, ygrowpt = 0, wgrowpt = 0, hgrowpt = 0;

graf_mouse(M_OFF,0L);
form_center(cur_tree, &xdial, &ydial, &wdial, &hdial);
form_dial(0, xgrowpt, ygrowpt, wgrowpt, hgrowpt,
             xdial,   ydial,   wdial,   hdial);
form_dial(1, xgrowpt, ygrowpt, wgrowpt, hgrowpt,
             xdial,   ydial,   wdial,   hdial); 
objc_draw(cur_tree, R_TREE, MAX_DEPTH, xdial, ydial, wdial, hdial);
graf_mouse(M_ON,0L);
exitobj = form_do(cur_tree, defindx); 
graf_mouse(M_OFF,0L);
form_dial(2, xgrowpt, ygrowpt, wgrowpt, hgrowpt,
             xdial,   ydial,   wdial,   hdial); 
form_dial(3, xgrowpt, ygrowpt, wgrowpt, hgrowpt,
             xdial,   ydial,   wdial,   hdial); 
objst_reset(cur_tree, exitobj, SELECTED);
graf_mouse(M_ON,0L);
return(exitobj);
}

/************************************************************************
 *
 * Do the zero-out operation.  Find out about the volume config, load
 *   the FAT into a buffer, and call the ASM routines to zero-out 
 *   everything that's not a bad sector.  Write the zeroed FAT back to
 *   both the primary and backup FAT on disk, then zero the root dir.
 *   If a volume label needs to be added, force media change on the 
 *   device, so GEMDOS will throw away it's buffers before the Fcreate,
 *   then do the Fcreate to place the volume label.
 *
 ************************************************************************/

do_zero()
{
int  fhandle;

if(!(p_bpb = Getbpb(drv_code)))
     {        
     form_alert(1,err_getbpb_alert);
     return(0);
     }

num_sectors = (p_bpb->bpb_fsiz > p_bpb->bpb_rdlen) 
               ? p_bpb->bpb_fsiz : p_bpb->bpb_rdlen;

p_buffer = Malloc((long)(num_sectors << 9));
if (p_buffer == 0L)
     {        
     form_alert(1,err_malloc_alert);
     return(0);
     }

num_sectors = p_bpb->bpb_fsiz;
fst_sector = p_bpb->bpb_fatrec - num_sectors;
stat = Rwabs(0,p_buffer,num_sectors,fst_sector,drv_code);
if (stat == 0L)
     {
     fixfat(p_buffer,p_bpb);
     stat = Rwabs(1,p_buffer,num_sectors,fst_sector,drv_code);
     if (stat == 0L)
          {
          fst_sector += num_sectors;
          stat = Rwabs(1,p_buffer,num_sectors,fst_sector,drv_code);
          if (stat == 0L)
               {
               fst_sector += num_sectors;
               num_sectors = p_bpb->bpb_rdlen;
               counter = num_sectors << 6;
               p_work = p_buffer;
               while (counter--)               
                    *p_work++ = 0L;
               stat = Rwabs(1,p_buffer,num_sectors,fst_sector,drv_code);
               }
          }
     }

Rwabs(0,0L,2,0,drv_code);    /* uh oh, here we go again - force mediach */
dmysearch[0] = 'A' + drv_code;
Fsfirst(dmysearch,0);

if (stat != 0L)
     {
     form_alert(1,err_ioerr_alert);
     return((int)stat);
     }

if (vol_label[0])
     {
     strcpy(wrk_path,vol_path);
     wrk_path[0] = drv_char[6];
     strcpy(&wrk_path[3],vol_label);
     fhandle = (int)Fcreate(wrk_path,0x0008);
     if (fhandle < 1)
          {
          form_alert(1,err_label_alert);
          return(1);
          }
     Fclose(fhandle);
     }

form_alert(1,ok_alldone_alert);
return(0);
}

/************************************************************************
 *
 * Main driver.
 *   Call init processing, loop through displaying the main dialog box.
 *   When a request to zero a device is selected, check for a volume 
 *   label on the disk.  If found, hide the dbox text encouraging the
 *   user to label the disk, and show the text stating that the label
 *   may be changed.  If there is no label, hide the change-able text,
 *   and unhide the text telling the user what a neat concept volume
 *   labels are.  If the user decides he really wants to zero the volume,
 *   invoke the routine to do that, then loop back to the main dbox.
 *
 ************************************************************************/

main()
{
int  selection;

prg_init();

while (1)
     {
     selection = dial_do(maintree,0);
     switch (selection)
          {
          case MAINBXEX:
               prg_exit();
          case MAINBXIN:
               dial_do(infotree,0);
               break;
          default:
               drv_code = selection - MAINBXA;
               search[0] = drv_char[6] = 'A' + drv_code;
               objfl_set(warntree,WARNCHLT,HIDETREE);
               objfl_set(warntree,WARNNOLT,HIDETREE);
               save_dta = Fgetdta();
               Fsetdta(&fs_dta);
               if (0L == Fsfirst(search,0x0008))
                    {
                    strcpy(vol_label,fs_dta.fs_name);
                    objfl_reset(warntree,WARNCHLT,HIDETREE);
                    }
               else
                    {
                    vol_label[0] = 0x00;
                    objfl_reset(warntree,WARNNOLT,HIDETREE);
                    }
               Fsetdta(save_dta);
               selection = dial_do(warntree,0);
               if (selection == WARNBXOK)
                    do_zero();
               break;
          }
     }
}
