********************************************************************
* maccel							****
* proportional mouse accelerator				****
* Copyright (c) 1989, Atari Corporation				****
*								****
* Permission is granted to copy this program, its source code	****
* and documentation file, provided they are kept together in	****
* an archive file exactly as they were originally distributed.	****
* 								****
********************************************************************
****
**** 890117 kbad	added 2-speed switch and config. message
**** 890116 kbad	Cleaned up for PD release.  Happy birthday, Dr. King
**** 881020 kbad	Hack hack hack.
****
*****************************************************************

****
* OS equates
****

.include atari			; Standard MADMAC Atari include file

kbs10		=	$e1b	; kbshift var for TOS 1.0
**** cursor modes
blink		=	$20000	; function codes as longwords
solid		=	$30000	; for speed & size

**** IKBD vector table
	.ABS
midivec:	ds.l	1
vkbderr:	ds.l	1
vmiderr:	ds.l	1
statvec:	ds.l	1
mousevec:	ds.l	1
clockvec:	ds.l	1
joyvec:		ds.l	1
midisys:	ds.l	1
ikbdsys:	ds.l	1

**** IKBD mouse packet offsets
	.ABS
button:		ds.b	1
xdelta:		ds.b	1
ydelta:		ds.b	1


****
* Some handy macros
****

.macro IPL7
	move	sr,-(sp)
	ori	#$700,sr
.endm

.macro IPL_RESTORE
	move	(sp)+,sr
.endm

.macro	Cconin
	Gemdos 1,2
.endm

*****************************************************************

	.TEXT
start:
	bra	startup			; jump past the handlers
****
* Here be the mouse packet handler...
****
* First, check the "pass" flag to see if the user has disabled acceleration.
* Then get the number of 200Hz ticks since the last mouse event.
* Two types of acceleration are provided, "slow" and "fast", determined
* by the "linear" flag.  "Slow" acceleration multiplies the X and Y deltas
* by 2 if only 0 or 1 tick has passed since the last event.  If more than
* 1 tick has passed, the deltas are left unchanged.  "Fast" acceleration
* is similar, but multiplies the deltas by 4 if 0 ticks have passed.
* Last, check to see if the user is Control-Shift-Alt-clicking to configure
* accelerator speed or enable/disable the accelerator.
****
* input:	pointer to mouse packet in a0
* output:	munged mouse packet if the mouse was moved fast
* uses:		d0-d2/a1
****
mymouse:
	move.l	oldmvec(pc),-(sp)	; so we can rts when we're done
	move.b	pass(pc),d0		; see if we should do anything
	bne.b	lbut			; pass flag is non-zero, skip accel
	
	move.b	xdelta(a0),d0		; get the delta x byte
	ext	d0			; and make it a word
	move.b	ydelta(a0),d1		; get the delta y byte
	ext	d1			; "WORD"
	move.l	_hz_200.w,d2		; what time is it now?
	sub.l	oldticks(pc),d2		; how many ticks has it been?
	move.l	_hz_200.w,oldticks	; save current time

	cmp.l	#1,d2			; check the number of 200Hz ticks
	bgt.b	cont			; > 1? leave it alone
	blt.b	fast			; = 0 make it go way fast
slow:	add	d0,d0			; = 1 make it go a little fast
	add	d1,d1
	bra.b	cont
fast:	move.b	linear(pc),d2		; check the linear flag
	bne.b	slow			; if linear, it only goes a little fast
	lsl	#2,d0			; otherwise, warp out
	lsl	#2,d1

cont:
	move.b	d0,xdelta(a0)		; put back delta x byte
	move.b	d1,ydelta(a0)		; and delta y byte

**** check for Control-Shift-Alt click

rbut:	btst	#0,(a0)			; right button down?
	bne.b	rstill			; yes, see if it was down before
	clr.b	rbdown			; button not down, clear flag
	bra.b	lbut			; and check left button
rstill:	tst.b	rbdown			; was it down before?
	bne.b	lbut			; yes, go check left button
	st	rbdown			; set right button flag
	bsr.b	ckcas			; check keys
	bne.b	out			; CAS not down, skip left button check
	eor.b	#1,linear		; toggle linear acceleration flag
lbut:	btst	#1,(a0)			; left button down?
	bne.b	lstill			; yes, see if it was down before
	clr.b	lbdown			; button not down, clear flag
	bra.b	out			; and scram
lstill:	tst.b	lbdown			; was it down before?
	bne.b	out			; yes, scram
	st	lbdown			; set button flag
	bsr.b	ckcas
	bne.b	out
	eor.b	#1,pass			; toggle pass flag
out:	rts

**** Look for Control-Alt-Shift, return with CCR Z flag set if they're down
ckcas:	move.l	pkbshift(pc),a1
	move.b	(a1),d1
	and.b	#$e,d1
	cmp.b	#$e,d1
	rts

**** Resident variables used by keyboard handler

oldmvec:	dc.l	0
oldticks:	dc.l	0
pkbshift:	dc.l	0
cookie:		dc.l	$AA006E
cookptr:	dc.l	0
pass:		dc.b	0
linear:		dc.b	0
lbdown:		dc.b	0
rbdown:		dc.b	0

****
* Here be the VBI routine that ensures we keep our vector
* ( the VDI will snarf it out from under us if we aren't looking! )
****

myvbl:
	move.l	mvloc(pc),a0		; get mouse vector
	move.l	(a0),a1			; for speed
	cmp.l	#mymouse,a1		; still ours?
	beq.b	endvbl			; yup, punt
	move.l	vdivec(pc),d0		; have we stored the VDI mouse vector?
	bne.b	testit			; yes, see who's going for it this time
	move.l	a1,vdivec		; store the VDI mouse vector
	bra.b	stealit			; and replace it
testit:
	cmpa.l	d0,a1			; is it VDI?
	bne.b	endvbl			; no, leave it be
stealit:
	move.l	a1,oldmvec
	move.l	#mymouse,(a0)
endvbl:
	rts

**** Resident variables used by VBI

mvloc:		dc.l	0
vdivec:		dc.l	0

resend:					; end of resident code segment

****
* Here be the initialization code
****

startup:

**** Warn the user that we're searching...

	lea	solid,a0
	bsr	Cursor
	lea	search(pc),a0
	bsr	WriteString

**** Get IKBD mouse vector and some system variables.

	Kbdvbase
	move.l	d0,a6
	lea	mousevec(a6),a6		; save the mouse vector loc
	move.l	a6,mvloc
	move.l	(a6),oldmvec		; save the mouse vector

	Super				; we're lookin' at low memory...

**** stuff the kbshift pointer

	move.l	_sysbase.w,a0
	move.l	(a0),d0			; check version number
	cmp	#$102,d0
	bge.b	not10
	move.l	#kbs10,pkbshift		; TOS 1.0 kbshift location
	bra.b	cookme
not10:	move.l	$24(a0),pkbshift	; other TOS's kbshift location

****
* Look through memory for the cookie to see if we're already installed.
* WARNING: this code is pretty psychotic.  But it's a bit faster than
*	your basic add loop.
****

cookme:	move.l	$14(a0),a0		; the_magic
	move.l	4(a0),a5		; A5 = endaesbss
	move.l	_memtop.w,d7		; D7 = _memtop
	sub.l	a5,d7			; bytes to check
	lsr.l	#1,d7			; / 2 = words to check
	divu	#$7fff,d7		; divide by 32k for # banks to check
	move.l	cookie(pc),d6		; D6 = cookie
**** Don't decrement d2 for the dbra count: check past memtop into screen.
**** Oh well.
dobank:	move	#$7fff,d0		; load a bank counter
more:	addq	#2,a5			; increment the pointer
	cmp.l	(a5),d6			; got us a cookie?
	dbeq	d0,more			; drop out if done looping or
	beq.b	isit			; if we found the cookie
**** Waste some time to show the user that we're still looking,
**** by using the dreaded "Townsend Effect"...
	btst	#0,d7			; even bank
	beq.b	cback			; cursor back
	lea	vt_fwd(pc),a0
	bsr.b	WriteString		; else cursor forward
	bra.b	nxbank
cback:	lea	vt_back(pc),a0
	bsr.b	WriteString
nxbank:	dbra	d7,dobank		; bump bank counter
	bra.b	setvex			; No cookies today!
isit:	cmp.l	4(a5),a5		; check the cookptr
	bne.b	more			; no pointer, jump back

itis:	User				; We're already installed, so...
	lea	instruc(pc),a0		; let the user know,
	bsr.b	WriteString
	lea	bad(pc),a0
	bsr.b	WriteString
	tst.b	8(a5)			; tell user if it's off or on,
	beq.b	sayon
	lea	offmsg(pc),a0
	bra.b	doomsg
sayon:	lea	onmsg(pc),a0	
doomsg:	bsr.b	WriteString
	lea	spdmsg(pc),a0
	bsr.b	WriteString
	tst.b	9(a5)			; tell user if it's fast or slow
	beq.b	sayfas
	lea	slomsg(pc),a0
	bra.b	dosmsg
sayfas:	lea	fasmsg(pc),a0
dosmsg:	bsr.b	WriteString
	lea	presmsg(pc),a0
	bsr.b	WriteString
	lea	blink,a0
	bsr.b	Cursor
	Cconin				; wait for a keypress,
	lea	vt_off(pc),a0
	bsr.b	WriteString
	Pterm0				; and go away.

****
* coupla subroutines...
****

WriteString:	pea	(a0)
		Gemdos 9,6
		rts

Cursor:		pea	(a0)
		Xbios 21,6
		rts

****
* set our vectors (at last!)
****

setvex:
	move.l	_hz_200.w,oldticks	; save the current time
	move.l	#cookie,cookptr		; set up the cookie pointer.

	IPL7				; p-p-pleeeze don't pull my ears!

	move.l	#mymouse,(a6)		; steal the mouse vector

****
* VBI installation code
* NOTE: the following does NOT handle extending the VBL queue past 8 entries
*	But since this is an auto folder program, it shouldn't matter.
****
	move.l	_vblqueue.w,a0		; get the queue and skip GEM's entry
nextvb: addq	#4,a0
	tst.l	(a0)			; nil entry?
	bne.b	nextvb			; no, try again
	move.l	#myvbl,(a0)		; yeah, grab it

	IPL_RESTORE			; Houston, we're go for interrupts

**** tell the nice user we succeeded
	lea	success(pc),a0
	bsr.b	WriteString
	lea	vt_off(pc),a0		; restore default cursor settings
	bsr.b	WriteString
	lea	blink,a0
	bsr.b	Cursor
	

**** wait one second for the message to sink in
	move.l	_hz_200.w,d0
	add.l	#200,d0
wait:	cmp.l	_hz_200.w,d0
	bge.b	wait

	User

**** Here we save ONLY the memory needed for the mouse and VBI handlers.
**** The rest of the code (everything after resend) goes away bye-bye.

	Ptermres #(resend-start+$100), #0 ; save handlers + basepage

****
* A buncha messages
****

search:	dc.b	"\r\n Mouse accelerator: Searching...."
****		cursor back 2
	dc.b	$1b,"D",$1b,"D"
****		cursor ON
vt_on:	dc.b	$1b,"e",0
****		cursor OFF
success:dc.b	$1b,"f"
	dc.b	$1b,"p\r    Mouse accelerator installed     \r\n",$1b,"q"
	dc.b	"   Copyright ",$bd," 1989, Atari Corp.\r\n"
instruc:dc.b	$1b,"p\r Control-Shift-Alternate-LeftClick  \r\n"
	dc.b	"  toggles accelerator on and off.   \r\n"
	dc.b	" Control-Shift-Alternate-RightClick \r\n"
	dc.b	"    toggles accelerator speed.      \r\n"
	dc.b	$1b,"q",0
bad:	dc.b	"   Accelerator already installed!   \r\n"
	dc.b	$1b,"p acceleration: ",$1b,"q",0
onmsg:	dc.b	"ON",$1b,"p ",0
offmsg:	dc.b	"OFF",$1b,"p",0
spdmsg:	dc.b	" speed: ",$1b,"q",0
fasmsg:	dc.b	"FAST",0
slomsg:	dc.b	"SLOW",0
presmsg:dc.b	$1b,"p      \r\n                   press a key...   "
	dc.b	$1b,"q"
vt_back:dc.b	$1b,"D",0
vt_off:	dc.b	$1b,"f",0
vt_fwd:	dc.b	$1b,"C",0

*the
.END
