/****************************************************************************/
/*                                                                          */
/* Module:     getsec.c - Get sectors from a disk into a file               */
/*                                                                          */
/* Programmer: George R. Woodside                                           */
/*                                                                          */
/* Date:       May 2, 1987                                                  */
/*                                                                          */
/****************************************************************************/

#include <stdio.h>
#include <osbind.h>
#include <ctype.h>
#include <errno.h>

extern  int   optind;                   /* index of unparsed argument       */
extern  char  *optarg;                  /* pointer to argument              */
extern  char  *sys_error();             /* error message function           */

unsigned  char  *buffer;                /* data buffer                      */

main(argc,argv)
int   argc;                             /* incoming argument count          */
char  *argv[];                          /* command line arguments           */

{
  register  int  drive  = -1;           /* designated drive                 */
  register  int  sstart = -1;           /* current track                    */
  register  int  send   = -1;           /* current side                     */

  register  long reply;                 /* I/O return code                  */

  unsigned  int  c;                     /* data character                   */

  register  char  *fname = (char *)NULL;/* file name                        */

  FILE  *fp;                            /* data file pointer                */

  if(argc == 1)                         /* if no arguments,                 */
    {
      printf("Usage: getsec -d drive -s start -e end -f file\n"); /* print usage display */
      printf("              -d     Drive to read (A/B/C...etc.\n");
      printf("              -s     Starting sector (from zero).\n");
      printf("              -e     Ending sector.\n");
      printf("              -f     File name.\n");
      exit(1);                          /* now punt                         */
    }                                   /* end no arguments                 */

  while( (c=getopt(argc,argv,"D:E:F:S:d:e:f:s:")) != EOF) /* parse options  */
    {
      switch(c)
        {
          case 'D':                     /* drive name                       */
          case 'd':
            c = *optarg;                /* get data character               */
            if(isupper(c))              /* if in upper case,                */
              c = tolower(c);           /* change to lower case             */
            drive = c - 'a';            /* set it                           */
            break;

          case 'E':                     /* ending sector                    */
          case 'e':
            send = atoi(optarg);        /* get ending sector number         */
            break;

          case 'F':                     /* file name                        */
          case 'f':
            fname = optarg;             /* save the name                    */
            break;

          case 'S':                     /* starting sector                  */
          case 's':
            sstart = atoi(optarg);      /* get ending sector number         */
            break;

        }                               /* end switch                       */
    }                                   /* end all options                  */

  if(optind != argc)                    /* if an unparsed argument,         */
    fname = argv[optind];               /* save the name                    */

  if(drive == -1)                       /* if no valid drive,               */
    {
      printf("You must specify a drive name.\n"); /* display nasty note     */
      exit(1);
    }                                   /* end no drive                     */

  if(sstart == -1)                      /* if no starting sector,           */
    {
      printf("You must specify a starting sector number.\n"); /* display nasty note */
      exit(1);
    }                                   /* end no start                     */

  if(send == -1)                        /* if no ending sector,             */
    {
      printf("You must specify an ending sector number.\n"); /* display nasty note */
      exit(1);
    }                                   /* end no end                       */

  if(send < sstart)                     /* if sector numbers bungled,       */
    {
      printf("Starting sector exceeds ending sector.\n"); /* display nasty note */
      exit(1);
    }                                   /* end bad numbers                  */

  if(fname == (char *)NULL)             /* if no file name,                 */
    {
      printf("You must specify a file name.\n"); /* display nasty note      */
      exit(1);
    }                                   /* end no file name                 */

  c = (send - sstart)+1;                /* count sectors                    */

  buffer = malloc(c * 512);             /* allocate buffer                  */

  if(buffer == (char *)NULL)            /* if malloc failed,                */
    {
      printf("Unable to allocate enough memory for buffer.\n"); /* say so,  */
      exit(1);                          /* and punt                         */
    }                                   /* end malloc failed                */

  reply = Rwabs(0,buffer,c,sstart,drive); /* read it                        */

  if(reply == AE_CHNG)                  /* if media change                  */
    {
      Getbpb(drive);                    /* clear media change               */
      reply = Rwabs(0,buffer,c,sstart,drive); /* retry the read             */
    }                                   /* end media change                 */

  if(reply)                             /* if a read error,                 */
    {
      printf("Error %ld reading data:",reply); /* show error                */
      printf(" %s\n",sys_error(reply)); /* now the text version             */
      exit(1);                          /* give up now                      */
    }                                   /* end read error                   */

  if((fp = fopen(fname,"wb")) == NULL)  /* if open fails,                   */
    {
      printf("Error opening file %s.\n",fname); /* log file name            */
      exit(1);                          /* and punt                         */
    }                                   /* end open error                   */

  c *= 512;                             /* change sectors to bytes          */

  if((fwrite(buffer,1,c,fp)) != c)      /* if write error,                  */
    printf("Error writing data file.\n"); /* nothing else we can do         */

  if((fclose(fp)) ==EOF)                /* if error on close,               */
    printf("Error closing data file.\n"); /* hopless case, here             */
}                                       /* end main                         */
