;**************************************************************************
;
;   QSTSM.S - small version of Quick ST, monochrome, AUTO folder version
;
;   (c) 1987, 1988, 1989 Darek Mihocka and Ignac Kolenko
;
;   last update: 08/11/89 17:00
;
;   This code is not public domain and may not be sold, altered, or
;   used in whole or part in any other software package without
;   permission from the authors.
;
;   This code can be assembled using the MADMAC macro assembler.
;   It must be assembled directly into a .PRG file, so just type
;       -p qstsm
;
;   The include file XATARI.SH is a customized version of the standard
;   ATARI.S file that comes with MADMAC. We have added some extra
;   constants, such as the negative line A variables, which were not
;   included by Atari, although they are fully documented in S.A.L.A.D.
;   (Still Another Line A Document).
;
;   This is a generic monochrome only version of Quick ST. "What does
;   that mean" you ask? Well, the monochrome part should be obvious. Do
;   not run this version in color. By "generic", it means that it is
;   not optimized for any one resolution. It should just as well on a
;   640x400 screen as on a 704x480 screen or any other size of upgrade
;   such as the Overscan/Hyperscreen. There are some assumptions made.
;   The size of a scan line is assumed to be a multiple of 4 bytes, and
;   the VT52 character height is assumed to be 8 or 16. The code will
;   probably break if the resolution exceeds 2048x1024.
;
;   This code is derived from one of the Quick ST 1.4 streams. Several
;   versions of the source code are maintained. Color only and monochrome
;   only versions. Of these, there are "small" versions (this one) and
;   "large versions" which are fully optimized for standard or Hyperscreen
;   monitors. Then there are combined color and mono versions. And finally
;   AUTO folder and desk accessory versions. It's a mess! This was the
;   smallest and simplest version.
;
;   It is intended for those people who are curious about how this program
;   works, or are maybe planning to write their own TSR and don't know
;   where to start. And there are a few programs which cause Quick ST to
;   break (such as the FSELXX.PRG programs) of which the authors are
;   unknown, and so we can't contact them to figure out what's wrong.
;   If you are a developer and discover a compatibility problem with
;   Quick ST and one of your products, look this over to see what's wrong.
;   In most cases it has turned out to be a conflict over stealing TRAP
;   vectors. Do NOT assume that you are the only TSR in memory, as GDOS
;   and the Diablo emulator like to think. It was a real pain to make
;   Quick ST fool these programs into working.
;
;   So how does Quick ST work? Well, this code won't tell you how the
;   large versions of Quick ST work (which are up to 100% faster), or how
;   to do fast color operations, or show all the neat instances of self
;   modifying code, and other goodies (what do you expect for free!).
;   But it does show how to install a TSR, how to hook TRAP vectors, how
;   to hook GEM operations, and how to totally break with Atari GDOS <grin>.
;   Most of this knowledge was either learned the hard way, or squeezed out
;   of people like Ian Lepore, Keith Gerdes, Charles Johnson, or Julius
;   Oklamcak. Thanks guys.
;
;   Ok, so here's what happens. First you gotta find your basepage (well,
;   not really, but this is "nicer"). Then find a stack and shrink your
;   TPA down to the minimum size. Quick ST keeps all its variables in
;   the text segment so that they can be accessed using PC relative code.
;   So there's no DATA or BSS segment. Then it goes hooks the TRAP vectors
;   and terminates (and stays resident).
;
;   So then another AUTO program loads, and then accessories, and then
;   the desktop comes up. By then, GEM has installed and moved the TRAP #2
;   vector to a safe place where no one else can overwrite it. This allows
;   it to work with Easel ST without any special checking, and it also
;   doesn't have to be paranoid about GEM overwriting the TRAP #2 vector.
;   Desk accessories that steal TRAP #2 have to worry, but at least they're
;   first in line to service the call.
;
;   The rest is pretty straight forward. The TRAP #13 handler looks for
;   Bconout calls and does VT52 text. The TRAP #1 handler looks for Cconws
;   calls and just calls the VT52 routines in a loop. The TRAP #2 handler
;   looks for VDI calls, and for a -1 instruction that GDOS uses to hook in.
;   It's a mess what it does with GDOS. Just make sure that GDOS loads before
;   Quick ST and all will be fine (usually).
;
;   Well, that's about it for now. If you discover any brain dead code, or
;   some piece of code that can be optimized, let us know!
;
;*************************************************************************/

    .include "xatari.sh"

    .macro setmode label
        move.w #\label-outB-2,outB+2
        move.w #\label-outC-2,outC+2
    .endm

    .macro setplot label
        move.w #\label-outCON1-2,outCON1+2
    .endm

    .macro setplots label
        move.w #\label-outCON2-2,outCON2+2
    .endm

    ; start of code!!

    .text

__main::
    lea     stack+512(pc),sp    ; load initial stack pointer
    lea     __main-256(pc),a2   ; a2 = basepage address

    move.l  codelen(a2),d0      ; compute total amount of memory req'd
    addi.l  #$100,d0            ; add basepage size
    move.l  d0,-(a7)            ; total memory used
    Mshrink a2,d0

    bsr    init                 ; initialization

; print title screen and terminate

    Puts    #szTitle
    move.l  (sp)+,d0
    Ptermres d0,#0


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   program variables, kept in .text for PC-rel access
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

stack:
    dcb.b 512,0

szTitle:
    dc.b $1b,'E',7,10
    dc.b 'Quick ST 1.48 Mono, Aug 11/89',13,10,10
    dc.b '(C) 1989 Darek Mihocka and Ignac Kolenko', 13, 10
    dc.b 0
    .even

plineTEXT:
    dc.l 0          ; pointer to start of current line that cursor is on

pLineA:
    dc.w 0          ; pointer to Line A structure (must be 16 bits!!!!)
pVW:
    dc.w 0          ; pointer to desktop virtual workstation

fGplus:
    dc.w 0          ; is G+PLUS loaded?
lGplusBD:
    dc.l 0          ; G+PLUS back door structure
fGDOS:
    dc.w $FFFF      ; is any GDOS installed? defualt is to assume yes

monocolmask:
    dc.l    $00000000

xmincl:
    dc.w    0
xmaxcl:
    dc.w    0
ymincl:
    dc.w    0
ymaxcl:
    dc.w    0

lLineStyle:
    dc.w 0
wLineStyle:
    dc.w 0          ; line style for v_pline

wShiftBits:
    dc.w 0          ; in v_gtext, # of bits out of alignment

vdiROMvec::
    dc.l 0

_OS_ver::
    dc.w    0       ; OS version (e.g. $0104)
_OS_date::          ; OS date (e.g. $06251985)
    dc.l    0

qcontrl:
    dc.l    0
qintin:
    dc.l    0
qptsin:
    dc.l    0
qintout:
    dc.l    0
qptsout:
    dc.l    0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   initialization
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

init:
    ; initialize the font arrays

    dc.w    $A000       ; returns A1 = pointer to array of font headers
    move.l  a0,a2
    move.w  a2,pLineA

; install new Trap vectors and re-initialize VT52

    Super

    move.w  #$472B,d0   ; "G+"
    trap    #2
    cmp.w   #$472B,d0
    beq.s   .noGplus

    move.l  d0,lGplusBD     ; pointer to G+ backdoor structure
    st      fGplus

.noGplus:
    moveq   #$FF,d0
    trap    #2
    move.l  d0,vdiROMvec
    

    move.l  $84,oldTrap1+2
    move.l  $88,oldTrap2+2
    move.l  $B4,oldTrap13+2

    ; Diablo emulator fix, copy 12 bytes right before trap #13 handler
    move.l  $B4,a0
    lea     myTrap13(pc),a1
    move.l  -(a0),-(a1)
    move.l  -(a0),-(a1)
    move.l  -(a0),-(a1)

    ; fill trap #2 jump table with pointer to old trap #2 vector
    lea     mpTrap2(pc),a1
    move.l  $88,a0
    move.w  #143,d0
.trap2:
    move.l  a0,(a1)+
    dbf     d0,.trap2

    lea     mpTrap2(pc),a1
    move.l  #v_ESC,20(a1)       ; command 5
    move.l  #vswr_mode,128(a1)  ; command 32
    move.l  #vsl_color,60(a1)   ; command 17
    move.l  #vsl_type,60(a1)    ; command 15
    move.l  #vst_color,88(a1)   ; command 22
    move.l  #vsf_color,100(a1)  ; command 25
    move.l  #v_gtext,32(a1)     ; command 8
    move.l  #v_pline,24(a1)     ; command 6
    move.l  #v_rectfl,456(a1)   ; command 114
    move.l  #vsl_udsty,452(a1)  ; command 113

    move.l  #myTrap1,$84
    move.l  #myTrap2,$88
    move.l  #myTrap13,$B4

    move.l  $44E,A1
    move.w  pLineA(pc),a2
    move.l  A1,V_CUR_AD(a2)
    move.l  A1,plineTEXT

    clr.w   V_CUR_X(a2)
    clr.w   V_CUR_Y(a2)
    clr.w   V_SAV_X(a2)
    clr.w   V_SAV_Y(a2)
    move.w  #1,V_COL_FG(a2)     ; set foreground = black
    clr.w   V_COL_BG(a2)        ; clear background color
    move.w  #2,V_HID_CNT(a2)    ; hide cursor since we are in a .PRG or .ACC
    bclr    #4,V_CURSTAT(a2)    ; clear inverse mode
    bclr    #3,V_CURSTAT(a2)    ; clear overflow mode
    bclr    #2,V_CURSTAT(a2)    ; disable cursor
    bclr    #1,V_CURSTAT(a2)    ; clear cursor phase
    bset    #0,V_CURSTAT(a2)    ; set flashing cursor

    ; get TOS version number
    move.l  $4F2,a0
    move.w  2(a0),_OS_ver
    move.l  $18(a0),_OS_date

    cmpi.w  #$0100,_OS_ver      ; is it TOS 1.0
    bne.s   .not10

    move.l  $0070,oldVBI+2      ; install VBI handler for cursor flash
    move.l  #myVBI,$0070

.not10:
    User
    rts



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   trap handlers
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; GEMDOS trap handler

myTrap1:
    move    usp,A0
    btst    #5,(sp)         ; check SR S bit
    beq.s   .1
    lea     6(sp),A0        ; pointer to parameters (if from S mode)
.1:
    cmp.w   #9,(a0)         ; check if GEMDOS #9 (Cconws)
    beq.s   myCconws

    cmp.w   #2,(a0)         ; check if GEMDOS #2 (Cconout)
    beq.s   myCc2

    move    (sp),ccr        ; preserve condition codes? why? I dunno.

oldTrap1:
    jmp     $ABBAABBA


myCconws:
    move.l  a3,-(sp)        ; save A3
    move.l  2(a0),A3        ; get string
    move.w  pLineA(pc),a2

    bsr     set_plot        ; just in case font changed!

    ; disable cursor
    addq.w  #1,V_HID_CNT(a2)
    bclr    #2,V_CURSTAT(a2)

myCc1:
    move.b  (a3)+,d0
    beq.s   xCcw
outC:
    bsr     outCON          ; self modifying line!
    move.b  (a3)+,d0
    bne.s   outC

xCcw:
    move.l  (sp)+,a3
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bset    #2,V_CURSTAT(a2)
.x:
    rte

myCc2:
    addq.l  #2,a0
myBc2:
    move.w  (a0),d0         ; get character
    move.w  pLineA(pc),a2

    addq.w  #1,V_HID_CNT(a2)

    ; plot character, self modifying line!!
outB:
    bsr     outCON

    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    rte

myBconout:
    move.w  (a0)+,d1        ; get device #
    cmpi.w  #CON,d1         ; is it CON device?
    beq.s   myBc2
    cmp.w   #RAWCON,d1      ; is it TEXT device?
    bne     oldTrap13
    move.w  (a0),d0         ; get character
    ext.w   d0
    move.w  pLineA(pc),a2

    ; disable cursor
    addq.w  #1,V_HID_CNT(a2)

    ; plot character - self modifying line!!
    move.w  #1,d1           ; force BGT
    bsr     outCON1         ; ya, so skip ESC handling

    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    rte

; BIOS trap handler

    ; Diablo Emulator fix, requires 12 bytes
    dc.l    0,0,0

myTrap13:
    move    usp,A0
    btst    #5,(sp)         ; check SR S bit
    beq.s   .1
    lea     6(sp),A0        ; pointer to parameters (if from S mode)

.1:
    cmp.w   #3,(a0)+        ; check if BIOS #3 (Bconout)
    beq   myBconout

oldTrap13:
    jmp     $ACDC0001


;
; TOS 1.0 cursor flash routine
;
; this is required because TOS 1.0 does't decrement the
; cursor flash counter when the cursor is not flashing.
; This causes the cursor to disappear when using Quick ST.
; This routine does what TOS 1.2 and 1.4 do in their VBIs.

myVBI:
    move.l  a2,-(sp)
    move.w  pLineA(pc),a2
    btst    #2,V_CURSTAT(a2)        ; is cursor on?
    beq.s   .x                      ; no, exit
    tst.w   V_HID_CNT(a2)           ; is cursor hidden
    bne.s   .x                      ; yes, exit
    btst    #0,V_CURSTAT(a2)        ; check flashing
    bne     .x                      ; yes, exit
    btst    #1,V_CURSTAT(a2)        ; chech cursor phase
    bne     .x                      ; on, exit
    subq.b  #1,V_CUR_CT(a2)         ; decrement cursor flash counter
    bne.s   .x                      ; not zero
    movem.l d0-d3/a0-a1,-(sp)
    bsr     set_cursor              ; flash cursor
    move.b  V_PERIOD(a2),V_CUR_CT(a2) ; reset counter
    movem.l (sp)+,d0-d3/a0-a1
.x:
    move.l  (sp)+,a2    
oldVBI:
    jmp     $94837564


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   console text output
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;
; outCON - output character to devices 2&5
;        - performs control and ESC character processing
;        - branches to appropriate text plotting routines
;
; D0 - character to print (already set)
; D1 - cursor X position (for post processing)
; D2 - cursor Y position (for post processing)
;
; A0 - pointer to screen char
; A1 - pointer to font
; A2 - pointer to lineA structure (already set)
;
; routines that don't need cursor post processing do an RTS
; otherwise they BRA new_cursor


outCON:
    ext.w   d0
    cmpi.b  #32,d0      ; if printable character, go do it

    ; self modifying line!!!
outCON1:
    bhi     set_plot    ; this will change at init time!!

    ; another self modifying line!!!
outCON2:
    beq     set_plot    ; this will change at init time!!

    cmpi.b  #27,d0
    bne.s   .ascii

    setmode vt_ESC
    bsr     set_plot
    rts

.ascii:
    movem.l d0/a0,-(sp)
    bsr     clear_cursor
    movem.l (sp)+,d0/a0
.7:
    cmpi.b  #7,d0
    blt     .noTEXT
    bne.s   .8

    btst    #2,$484
    beq     .noTEXT


    ; fake a Bconout (2,7)
    ; jump to old BIOS vector and return to RTS
    movem.l d3-d7/a0-a6,-(sp)
    move.l  #$00020007,-(sp)
    move.w  #$03,-(sp)
    pea     .7exit(pc)
    move.w  SR,-(sp)
    bra     oldTrap13   ; fake trap #13

.7exit:
    addq.l  #6,sp
    movem.l (sp)+,d3-d7/a0-a6
    rts

.8:
    cmpi.b  #8,d0
    bne.s   .9          ; branch if not BS

    move.w  V_CUR_X(a2),d1      ; if already 0, can't backspace
    beq.s   .noTEXT
    bsr.s   clear_cursor
    subq.w  #1,d1               ; cursor left
    move.w  V_CUR_Y(a2),d2
    bra     new_cursor

.9:
    cmpi.b  #9,d0
    bne.s   .10         ; branch if not TAB

.tab:
    moveq   #0,d0
    bsr     outCON2
    move.w  V_CUR_X(a2),d1
    cmp.w   V_CEL_MX(a2),d1
    beq.s   .x8
    andi.w  #$07,d1
    bne.s   .tab
.x8:
    rts

.10:
    cmpi.b  #10,d0      ; is it LF
    beq     vt_LF

.11:
    cmpi.b  #11,d0      ; is it VT? just do a LF
    beq     vt_LF

.12:
    cmpi.b  #12,d0
    beq     vt_LF

.13:
    cmpi.b  #13,d0
    beq     vt_CR

.noTEXT:
    rts


set_cursor:
    ; if cursor was off, invert the char
    bset     #1,V_CURSTAT(a2)    ; check cursor phase
    beq.s   toggle_cursor
    rts

clear_cursor:
    ; if cursor was on, invert the char
    bclr    #1,V_CURSTAT(a2)    ; check cursor phase
    bne.s   toggle_cursor
    rts

toggle_cursor:
    move.l  V_CUR_AD(a2),a0
    move.w  V_CEL_HT(a2),d0
    subq.w  #1,d0

.m:
    not.b   (a0)
    adda.w  BYTES_LIN(a2),a0
    dbf     d0,.m
    move.b  V_PERIOD(a2),V_CUR_CT(a2) ; reset counter
    rts


; after every ESC sequence or other memorable event, we re-check the
; lineA variables and set a new plotting routine just in case

set_plot:
    movem.w d1/d2,-(sp)
    move.w  V_COL_FG(a2),d1
    move.w  V_COL_BG(a2),d2
    btst    #4,V_CURSTAT(a2)
    beq.s   .s_p
    exg     d1,d2
.s_p:
    bsr.s   .sp
    movem.w (sp)+,d1/d2
    rts

.sp:
    setplot monogen
    setplots monogen

    cmp.w   d1,d2
    beq.s   .1genw
    tst     d2
    bne.s   .1geni
    rts

.1genw:
    tst     d2
    bne.s   .1genb
    setplot mspcgen
    rts

.1geni:
    setplot monogeni
    setplots mspcgeni
    rts

.1genb:
    setplot mspcgeni
    setplots mspcgeni
    rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   raw text output routines
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; monochrome plotting routine

monogen:
    move.l  V_CUR_AD(a2),a0
    move.l  V_FNT_AD(a2),a1    ; get pointer to font
    andi.w  #$00FF,d0       ; offset to char data in font
    adda.w  D0,A1           ; get address of char data in font
    move.w  V_CEL_HT(a2),d1

    not.w   d1
    andi.w  #31,d1      ; limit to 31 scan lines
    add.w   d1,d1
    move.w  d1,d0
    add.w   d1,d1
    add.w   d0,d1       ; multiply by 6

    move.w  BYTES_LIN(a2),d0
    move.w  V_FNT_WD(a2),d2
    jmp     .here(PC,d1.w)

.here:
    .rept   30
    move.b  (a1),(a0)
    adda.w  d2,a1       ; next scan line in font
    adda.w  d0,a0       ; next scan line on screen
    .endr
    move.b  (a1),(a0)

    move.l  V_CUR_AD(a2),a0
    addq.l  #1,a0
    bclr    #1,V_CURSTAT(a2)    ; clear cursor phase

    move.w  V_CUR_X(a2),d1      ; compare cursor position to max
    cmp.w   V_CEL_MX(a2),d1
    bcc.s   .25

    addq.w  #1,d1               ; else increment x and cursor address
    move.w  d1,V_CUR_X(a2)
    move.l  a0,V_CUR_AD(a2)     ; store new position of cursor
.25x:
    rts

.25:
    btst    #3,V_CURSTAT(a2)    ; check overflow bit
    beq.s   .25x
    move.w  V_CUR_Y(a2),d2
    bra     vt_CRLF             ; if set, go do CRLF, and RTS
    rts

monogeni:
    move.l  V_CUR_AD(a2),a0
    move.l  V_FNT_AD(a2),a1    ; get pointer to font
    andi.w  #$00FF,d0       ; offset to char data in font
    adda.w  D0,A1           ; get address of char data in font
    move.w  V_CEL_HT(a2),d1

    not.w   d1
    andi.w  #31,d1      ; limit to 31 scan lines
    add.w   d1,d1
    move.w  d1,d0
    add.w   d1,d1
    add.w   d1,d1
    add.w   d0,d1       ; multiply by 10

    move.w  BYTES_LIN(a2),d0
    move.w  V_FNT_WD(a2),d2
    jmp     .herei(PC,d1.w)

.herei:
    .rept   30
    move.b  (a1),d1
    not     d1
    move.b  d1,(a0)
    adda.w  d2,a1       ; next scan line in font
    adda.w  d0,a0       ; next scan line on screen
    .endr
    move.b  (a1),(a0)
    not.b   (a0)

    move.l  V_CUR_AD(a2),a0
    addq.l  #1,a0
    bclr    #1,V_CURSTAT(a2)    ; clear cursor phase

    move.w  V_CUR_X(a2),d1      ; compare cursor position to max
    cmp.w   V_CEL_MX(a2),d1
    bcc.s   .25

    addq.w  #1,d1               ; else increment x and cursor address
    move.w  d1,V_CUR_X(a2)
    move.l  a0,V_CUR_AD(a2)     ; store new position of cursor
.25x:
    rts

.25:
    btst    #3,V_CURSTAT(a2)    ; check overflow bit
    beq.s   .25x
    move.w  V_CUR_Y(a2),d2
    bra     vt_CRLF             ; if set, go do CRLF, and RTS
    rts

mspcgen:
    move.l  V_CUR_AD(a2),a0
    move.w  V_CEL_HT(a2),d1

    not.w   d1
    andi.w  #31,d1      ; limit to 31 scan lines
    add.w   d1,d1
    add.w   d1,d1       ; multiply by 4

    move.w  BYTES_LIN(a2),d0
    jmp     .here(PC,d1.w)

.here:
    .rept   30
    clr.b   (a0)
    adda.w  d0,a0       ; next scan line on screen
    .endr
    clr.b   (a0)

    move.l  V_CUR_AD(a2),a0
    addq.l  #1,a0
    bclr    #1,V_CURSTAT(a2)    ; clear cursor phase

    move.w  V_CUR_X(a2),d1      ; compare cursor position to max
    cmp.w   V_CEL_MX(a2),d1
    bcc.s   .25

    addq.w  #1,d1               ; else increment x and cursor address
    move.w  d1,V_CUR_X(a2)
    move.l  a0,V_CUR_AD(a2)     ; store new position of cursor
.25x:
    rts

.25:
    btst    #3,V_CURSTAT(a2)    ; check overflow bit
    beq.s   .25x
    move.w  V_CUR_Y(a2),d2
    bra     vt_CRLF             ; if set, go do CRLF, and RTS
    rts

mspcgeni:
    move.l  V_CUR_AD(a2),a0
    move.w  V_CEL_HT(a2),d1

    not.w   d1
    andi.w  #31,d1      ; limit to 31 scan lines
    add.w   d1,d1
    add.w   d1,d1       ; multiply by 4

    move.w  BYTES_LIN(a2),d0
    moveq   #$FF,d2
    jmp     .here(PC,d1.w)

.here:
    .rept   30
    move.b  d2,(a0)
    adda.w  d0,a0       ; next scan line on screen
    .endr
    move.b  d2,(a0)

    move.l  V_CUR_AD(a2),a0
    addq.l  #1,a0
    bclr    #1,V_CURSTAT(a2)    ; clear cursor phase

    move.w  V_CUR_X(a2),d1      ; compare cursor position to max
    cmp.w   V_CEL_MX(a2),d1
    bcc.s   .25

    addq.w  #1,d1               ; else increment x and cursor address
    move.w  d1,V_CUR_X(a2)
    move.l  a0,V_CUR_AD(a2)     ; store new position of cursor
.25x:
    rts

.25:
    btst    #3,V_CURSTAT(a2)    ; check overflow bit
    beq.s   .25x
    move.w  V_CUR_Y(a2),d2
    bra     vt_CRLF             ; if set, go do CRLF, and RTS
    rts


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   ESCape processing and vt_ routines
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ; Process escape code
    ; Use a jump table

vt_ESC:
    andi.w  #$00FF,d0       ; clear high byte
    subi.b  #'A',d0
    bcs     vt_ex

    cmpi.b  #'x'-'A',d0
    bcc     vt_ex

    add.w   d0,d0
    add.w   d0,d0
    move.l  vtjump(PC,d0.w),a0
    jmp     (a0)

vtjump:
    dc.l    vt_up       ; A
    dc.l    vt_down     ; B
    dc.l    vt_right    ; C
    dc.l    vt_left     ; D
    dc.l    vt_cls      ; E
    dc.l    vt_ex       ; F
    dc.l    vt_ex       ; G
    dc.l    vt_home     ; H
    dc.l    vt_ups      ; I
    dc.l    vt_ceos     ; J
    dc.l    vt_eeol     ; K
    dc.l    vt_ins_lin  ; L
    dc.l    vt_del_lin  ; M
    dc.l    vt_ex       ; N
    dc.l    vt_ex       ; O
    dc.l    vt_ex       ; P
    dc.l    vt_ex       ; Q
    dc.l    vt_ex       ; R
    dc.l    vt_ex       ; S
    dc.l    vt_ex       ; T
    dc.l    vt_ex       ; U
    dc.l    vt_ex       ; V
    dc.l    vt_ex       ; W
    dc.l    vt_ex       ; X
    dc.l    vt_pos_cursor1 ; Y
    dc.l    vt_ex       ; Z
    dc.l    vt_ex       ; !
    dc.l    vt_ex       ; !
    dc.l    vt_ex       ; !
    dc.l    vt_ex       ; !
    dc.l    vt_ex       ; !
    dc.l    vt_ex       ; @
    dc.l    vt_ex       ; a
    dc.l    vt_set_fg1  ; b
    dc.l    vt_set_bg1  ; c
    dc.l    vt_ctos     ; d
    dc.l    vt_on       ; e
    dc.l    vt_off      ; f
    dc.l    vt_ex       ; g
    dc.l    vt_ex       ; h
    dc.l    vt_ex       ; i
    dc.l    vt_save_cursor_pos ; j
    dc.l    vt_restore_cursor_pos ; k
    dc.l    vt_eline    ; l
    dc.l    vt_ex       ; m
    dc.l    vt_ex       ; n
    dc.l    vt_cbol     ; o
    dc.l    vt_set_inv  ; p
    dc.l    vt_clear_inv ; q
    dc.l    vt_ex       ; r
    dc.l    vt_ex       ; s
    dc.l    vt_ex       ; t
    dc.l    vt_ex       ; u
    dc.l    vt_set_wrap ; v
    dc.l    vt_clear_wrap ; w


; unknown escape sequence, ignore it

vt_ex:
    setmode outCON
    rts


; exit escape sequence that moved cursor

vt_ex_cur:
    bsr     new_curs_pos    ; D1,D2 must already be loaded with X,Y
    setmode outCON
    bsr     set_plot
    rts

;
;   ESC e - cursor on
;

vt_on:
    cmpi.w  #1,V_HID_CNT(a2)
    beq.s   .x
    move.w  #1,V_HID_CNT(a2)
    bset    #2,V_CURSTAT(a2)
.x:
    bra     vt_ex

;
;   ESC f - cursor off
;

vt_off:
    addq.w  #1,V_HID_CNT(a2)
    bclr    #2,V_CURSTAT(a2)
    bsr     clear_cursor
    bra     vt_ex


;
;   ESC A - cursor up, with scroll
;

vt_up:
    move.w  V_CUR_X(a2),d1
    move.w  V_CUR_Y(a2),d2
    beq     vt_ex_cur
    
    bsr     clear_cursor
    subq.w  #1,d2
    bra     vt_ex_cur

;
;   ESC B - cursor down
;

vt_down:
    move.w  V_CUR_X(a2),d1
    move.w  V_CUR_Y(a2),d2
    cmp.w   V_CEL_MY(a2),d2
    bcc     vt_ex_cur
    
    bsr     clear_cursor
    addq.w  #1,d2
    bra     vt_ex_cur

;
;   ESC C - cursor left
;

vt_left:
    move.w  V_CUR_Y(a2),d2
    move.w  V_CUR_X(a2),d1
    beq     vt_ex_cur
    
    bsr     clear_cursor
    subq.w  #1,d1
    bra     vt_ex_cur

;
;   ESC D - cursor right
;

vt_right:
    move.w  V_CUR_Y(a2),d2
    move.w  V_CUR_X(a2),d1
    cmp.w   V_CEL_MX(a2),d1
    bcc     vt_ex_cur
    
    bsr     clear_cursor
    addq.w  #1,d1
    bra     vt_ex_cur


;
;   ESC I - cursor up, with scroll
;

vt_ups:
    move.w  V_CUR_Y(a2),d2      ; if at top
    beq     vt_ins_lin          ; then insert line
    
    bsr     clear_cursor
    subq.w  #1,d2
    move.w  V_CUR_X(a2),d1
    bra     vt_ex_cur

;
;   ESC L - insert line
;

vt_ins_lin:
    bsr     set_plot

    bsr     clear_cursor

    movem.l d3/a3/a4,-(sp)

    move.l  V_CUR_AD(a2),a3
    move.l  plineTEXT(pc),a4
    move.w  V_CUR_X(a2),d3

    move.w  V_CEL_MY(a2),d0
    sub.w   V_CUR_Y(a2),d0      ; how many lines to scroll
    beq.s   .x                  ; none
    bcs.s   .x                  ; negative??

    mulu.w  V_CEL_HT(a2),d0     ; total number of scan lines to scroll
    move.w  BYTES_LIN(a2),d1    ; width of scan line in bytes
    mulu    d1,d0               ; total number of bytes to scroll
    move.l  d0,d1
    lsr.l   #4,d0               ; divide by 16
    subq.w  #1,d0

    move.l  a4,a0               ; pointer to current line
    move.l  a0,a1
    adda.w  V_CEL_WR(a2),a0     ; pointer to next line

    add.l   d1,a0
    add.l   d1,a1               ; go to and of screen and work back

.scroll:
    move.l  -(a1),-(a0)
    move.l  -(a1),-(a0)
    move.l  -(a1),-(a0)
    move.l  -(a1),-(a0)
    dbf     d0,.scroll

    move.l  a4,plineTEXT
    bsr     vt_eline            ; clear current line

.x:

    move.l  a4,V_CUR_AD(a2)
    move.l  a4,plineTEXT
    clr.w   V_CUR_X(a2)

    movem.l (sp)+,d3/a3/a4
    bra     vt_ex

;
;   ESC M - delete line
;

vt_del_lin:
    bsr     set_plot

    bsr     clear_cursor

    bsr     vt_eline            ; clear current line (in case at bottom)

    movem.l d3/a3/a4,-(sp)

    move.l  V_CUR_AD(a2),a3
    move.l  plineTEXT(pc),a4
    move.w  V_CUR_X(a2),d3

    move.w  V_CEL_MY(a2),d0
    sub.w   V_CUR_Y(a2),d0      ; how many lines to scroll
    beq.s   .x                  ; none
    bcs.s   .x                  ; negative??

    mulu.w  V_CEL_HT(a2),d0     ; total number of scan lines to scroll
    move.w  BYTES_LIN(a2),d1    ; width of scan line in bytes
    mulu    d1,d0               ; total number of bytes to scroll
    lsr.l   #4,d0               ; divide by 16
    subq.w  #1,d0

    move.l  a4,a0               ; pointer to current line
    move.l  a0,a1
    adda.w  V_CEL_WR(a2),a0     ; pointer to next line

.scroll:
    move.l  (a0)+,(a1)+
    move.l  (a0)+,(a1)+
    move.l  (a0)+,(a1)+
    move.l  (a0)+,(a1)+
    dbf     d0,.scroll

    move.l  a1,plineTEXT
    bsr     vt_eline            ; clear bottom line

.x:
    move.l  a4,V_CUR_AD(a2)
    move.l  a4,plineTEXT
    clr.w   V_CUR_X(a2)

    movem.l (sp)+,d3/a3/a4
    bra     vt_ex

;
;   ESC Y - reposition cursor
;

tempy:  dc.w    0

vt_pos_cursor1:
    setmode vt_pos_cursor2
    rts

vt_pos_cursor2:
    subi.w  #32,d0
    move.w  D0,tempy
    setmode vt_pos_cursor3
    rts

vt_pos_cursor3:
    subi.w  #32,d0
    move.w  D0,d1
    move.w  tempy(pc),d2

    bsr     clear_cursor

; limit X and Y to screen dimensions

    cmp.w   V_CEL_MX(a2),d1
    bcs.s   .11
    move.w  V_CEL_MX(a2),d1
.11:
    cmp.w   V_CEL_MY(a2),d2
    bcs.s   .12
    move.w  V_CEL_MY(a2),d2
.12:

    setmode outCON
    bsr     set_plot
    bra     new_curs_pos


vt_set_bg1:
    setmode vt_get_bg
    rts

vt_get_bg:
    andi.w  #1,d0
    move.w  d0,V_COL_BG(a2)

    btst    #0,d0
    sne     d2
    ext.w   d2
    ext.l   d2
    move.l  d2,monocolmask

    setmode outCON
    bra     set_plot

vt_set_fg1:
    setmode vt_get_fg
    rts

vt_get_fg:
    andi.w  #1,d0
    move.w  d0,V_COL_FG(a2)

    move.w  V_COL_BG(a2),d0
    btst    #0,d0
    sne     d2
    ext.w   d2
    ext.l   d2
    move.l  d2,monocolmask

    setmode outCON
    bra     set_plot

vt_set_wrap:
    bset    #3,V_CURSTAT(a2)
    setmode outCON
    bsr     set_plot
    rts

vt_clear_wrap:
    bclr    #3,V_CURSTAT(a2)
    setmode outCON
    bsr     set_plot
    rts

vt_set_inv:
    bset    #4,V_CURSTAT(a2)
    setmode outCON
    bra     set_plot

vt_clear_inv:
    bclr    #4,V_CURSTAT(a2)
    setmode outCON
    bra     set_plot


; Carriage Return
;
; clears cursor
; sets cursor X to 0
; sets cursor to start of current line

vt_CR:
    bsr     clear_cursor
    move.w  V_CUR_X(a2),d1
    move.w  V_CUR_Y(a2),d2
    bsr     new_curs_pos
    move.l  plineTEXT(pc),V_CUR_AD(a2)
    clr.w   V_CUR_X(a2)
    bsr     set_plot
    rts


vt_CRLF:
    bsr.s   vt_CR
    bra.s   vt_LF2

vt_LF:
    bsr     clear_cursor

vt_LF2:
    move.w  V_CUR_Y(a2),d2
    cmp.w   V_CEL_MY(a2),d2
    bcc.s   .go              ; if Y >= YMAX then must scroll

    ; move cursor down one line (cursor already cleared)
    move.w  V_CUR_X(a2),d1
    addq.w  #1,d2
    bra     new_cursor
    move.w  d2,V_CUR_Y(a2)
    move.l  V_CUR_AD(a2),a0
    adda.w  V_CEL_WR(a2),a0
    move.l  a0,V_CUR_AD(a2)
    move.l  plineTEXT(pc),a0
    adda.w  V_CEL_WR(a2),a0
    move.l  a0,plineTEXT
    bsr     set_plot
    rts

.go:
    move.w  V_CEL_MY(a2),V_CUR_Y(a2)
    move.l  $44E,A0     ; get pointer to screen
    adda.w  V_CUR_OF(a2),A0 ; offset to first char

    move.l  A0,A1
    adda.w  V_CEL_WR(a2),A1 ; offset to next line

    move.w  V_CEL_HT(a2),d0
    mulu    V_CEL_MY(a2),d0 ; total number of scan lines to scroll
    move.w  BYTES_LIN(a2),d1    ; width of scan line in bytes
    mulu    d1,d0
    lsr.w   #3,d1           ; divide by 8
    bcc     .160

    ; scroll in 80 byte blocks
    divu    #80,d0          ; total number of 80 byte blocks to scroll
    subq.w  #1,d0

    movem.l d3-d7/a3-a6,-(sp)

.scroll80:
    movem.l (a1)+,D2-D7/A3-A6
    movem.l D2-D7/A3-A6,(a0)
    movem.l (a1)+,D2-D7/A3-A6
    movem.l D2-D7/A3-A6,40(a0)
    lea     80(a0),a0
    dbf     D0,.scroll80

    movem.l (sp)+,d3-d7/a3-a6
    bra     .dobottom

.160:
    ; scroll in 160 byte blocks
    divu    #160,d0          ; total number of 160 byte blocks to scroll

    subq.w  #1,d0

    movem.l d1-d7/a2-a6,-(sp)

    cmpi.w  #2,d0
    bcs     .scroll160

.scroll320:
    movem.l (a1)+,D1-D7/A2-A6
    movem.l D1-D7/A2-A6,(a0)
    movem.l (a1)+,D1-D7/A2-A6
    movem.l D1-D7/A2-A6,48(a0)
    movem.l (a1)+,D1-D7/A2-A6
    movem.l D1-D7/A2-A6,96(a0)
    movem.l (a1)+,D1-D7/A2-A6
    movem.l D1-D7/A2-A6,144(a0)
    movem.l (a1)+,D1-D7/A2-A6
    movem.l D1-D7/A2-A6,192(a0)
    movem.l (a1)+,D1-D7/A2-A6
    movem.l D1-D7/A2-A6,240(a0)
    movem.l (a1)+,D1-D7/a2
    movem.l D1-D7/a2,288(a0)
    lea     320(a0),a0

    subq.w  #2,d0
    cmpi.w  #2,d0
    bcc     .scroll320

.scroll160:
    movem.l (a1)+,D2-D7/A3-A6
    movem.l D2-D7/A3-A6,(a0)
    movem.l (a1)+,D2-D7/A3-A6
    movem.l D2-D7/A3-A6,40(a0)
    movem.l (a1)+,D2-D7/A3-A6
    movem.l D2-D7/A3-A6,80(a0)
    movem.l (a1)+,D2-D7/A3-A6
    movem.l D2-D7/A3-A6,120(a0)
    lea     160(a0),a0
    dbf     D0,.scroll160

    movem.l (sp)+,d1-d7/a2-a6

.dobottom:
    move.w  V_CEL_WR(a2),d0
    lsr.w   #5,d0
    subq.w  #1,d0

    ; initialize D1 with proper background color
    move.l  monocolmask(pc),d1

.clear:
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    dbf     d0,.clear

    clr.w   d1
    move.w  V_CEL_MY(a2),d2

    bsr     set_plot

    ; fall through

new_cursor:
; this is the text out cleanup code
; compare the latest VT52 cursor position with what we last had.
; if they're different, calculate a new cursor address

    cmp.w    V_CUR_Y(a2),d2     ; did Y change?
    beq.s    x1

    ; force new X and Y, with checking of boundaries
new_curs_pos:
    cmp.w   V_CEL_MY(a2),d2 ; if below max Y
    bcs.s   .newy           ; then just update cursor
    move.w  V_CEL_MY(a2),d2 ; else set Y to bottom of screen

.newy:
    move.w  d1,V_CUR_X(a2)  ; save new X position
    move.w  d2,V_CUR_Y(a2)  ; save new Y position

    movea.l $44E,a0         ; top of screen
    adda.w  V_CUR_OF(a2),a0 ; add offset from top
    move.w  V_CEL_WR(a2),d0
    mulu.w  d2,d0           ; offset to scan line

    adda.l  d0,a0
    move.l  a0,plineTEXT    ; save new start of scan
    bra.s   newx

x1:
    cmp.w   V_CUR_X(a2),d1
    beq.s   exit

    move.l  plineTEXT(pc),a0

newx:
    cmp.w   V_CEL_MX(a2),d1 ; if below max X
    blt.s   .x2             ; then just update cursor

    move.w  V_CEL_MX(a2),d1
    
.x2:   
    move.w  d1,V_CUR_X(a2)  ; save new X position
    adda.w  d1,a0
    move.l  a0,V_CUR_AD(a2) ; save new cursor position

exit:
    rts


vt_save_cursor_pos:
    move.w  V_CUR_X(a2),V_SAV_X(a2)
    move.w  V_CUR_Y(a2),V_SAV_Y(a2)
    bset    #5,V_CURSTAT(a2)
    bsr     set_plot
    setmode outCON
    rts

vt_restore_cursor_pos:
    btst    #5,V_CURSTAT(a2)
    beq.s   vt_home     ; if not saved, just home

    bsr     clear_cursor
    move.w  V_SAV_X(a2),d1
    move.w  V_SAV_Y(a2),d2
    bclr    #5,V_CURSTAT(a2)
    bra     vt_ex_cur

vt_home:
    bsr     clear_cursor
    move.l  $44E,A0     ; get pointer to screen
    move.l  A0,V_CUR_AD(a2)
    move.l  a0,plineTEXT
    clr.w   V_CUR_X(a2)
    clr.w   V_CUR_Y(a2)        ; set cursor to (0,0)
    bra     vt_ex

;
; ESC E - clear screen
;

vt_cls:
    setmode outCON
    bsr     set_plot

    move.l  $44E,A0             ; get pointer to screen
    adda.w  V_CUR_OF(a2),A0     ; offset to first char
    move.l  A0,V_CUR_AD(a2)
    move.l  a0,plineTEXT
    clr.w   V_CUR_X(a2)
    clr.w   V_CUR_Y(a2)         ; set cursor to (0,0)
    bclr    #1,V_CURSTAT(a2)    ; clear cursor phase

    move.w  V_CEL_HT(a2),d0
    mulu    V_CEL_MY(a2),d0
    add.w   V_CEL_HT(a2),d0     ; total number of bytes to clear
    move.w  BYTES_LIN(a2),d1        ; width of scan line in bytes
    mulu    d1,d0               ; total number of bytes to scroll
    lsr.l   #4,d0               ; divide by 16
    subq.w  #1,d0

    move.l  monocolmask(pc),d1

.clear:
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    dbf     d0,.clear

    rts

;
; ESC J - clear from cursor to end of screen
;

vt_ceos:
    setmode outCON
    bsr     set_plot

    bsr     clear_cursor

    movem.l d3/a3/a4,-(sp)

    move.l  V_CUR_AD(a2),a3
    move.l  plineTEXT(pc),a4
    move.w  V_CUR_X(a2),d3

    bsr     vt_eeol             ; clear to end of line

    ; now do something similar to a cls expect not from start of screen
    move.w  V_CEL_MY(a2),d0
    sub.w   V_CUR_Y(a2),d0      ; how many lines to clear
    beq.s   .x                  ; none
    bcs.s   .x                  ; negative??

    mulu.w  V_CEL_HT(a2),d0     ; total number of scan lines to clear
    move.w  BYTES_LIN(a2),d1        ; width of scan line in bytes
    mulu    d1,d0               ; total number of bytes to scroll
    lsr.l   #4,d0               ; divide by 16
    subq.w  #1,d0

    move.l  a4,a0               ; pointer to current line
    adda.w  V_CEL_WR(a2),a0     ; pointer to next line
    move.w  BYTES_LIN(a2),d1        ; width of scan line in bytes

    ; initialize D1 with proper background color
    move.l  monocolmask(pc),d1

.clear:
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    dbf     d0,.clear

.x:
    move.l  a3,V_CUR_AD(a2)
    move.l  a4,plineTEXT
    move.w  d3,V_CUR_X(a2)

    movem.l (sp)+,d3/a3/a4
    rts
    
;
; ESC d - clear from top of screen to current cursor position
;

vt_ctos:
    setmode outCON
    bsr     set_plot

    bsr     clear_cursor

    movem.l d3/a3,-(sp)

    move.l  V_CUR_AD(a2),a3
    move.w  V_CUR_X(a2),d3     ; get x

    move.l  $44E,A0             ; get pointer to screen
    adda.w  V_CUR_OF(a2),A0     ; offset to first char
    move.l  plineTEXT(pc),a1

    move.l  monocolmask(pc),d1

.clear:
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    cmpa.l  a1,a0
    bcs     .clear

    movem.l (sp)+,d3/a3
    bsr     vt_cbol             ; clear from beginning of line to cursor
    rts

;
; ESC l - erase line
;

vt_eline:
    move.l  plineTEXT(pc),V_CUR_AD(a2)
    move.l  plineTEXT(pc),a0
    clr.w   V_CUR_X(a2)
    move.w  V_CEL_WR(a2),d0
    lsr.w   #4,d0
    subq.w  #1,d0
.1:
    clr.l   (a0)+
    clr.l   (a0)+
    clr.l   (a0)+
    clr.l   (a0)+
    dbf     d0,.1

    bclr    #1,V_CURSTAT(a2)    ; clear cursor phase
    setmode outCON
    bsr     set_plot
    rts

;
; ESC K - clear remainder of line, cursor position unchanged
;

vt_eeol:
    setmode outCON
    bsr     set_plot

    movem.l d3-d6/a3,-(sp)

    move.l  V_CUR_AD(a2),a3
    move.w  V_CUR_X(a2),d3     ; get x
    move.w  V_CEL_MX(a2),d4
    bclr    #3,V_CURSTAT(a2)
    sne     d5

.loop:
    clr.w   d0
    bsr     outCON2

    move.w  V_CUR_X(a2),d0
    cmp.w   d0,d4
    beq     .last

    andi.w  #3,d0
    bne.s   .loop

    ; X is a multiple of 4, and we're not at the end of the line yet
    ; so clear the end of the line using longs

.fast:
    ; D0 = number of words from start of scan line to skip to X
    ; A0 = pointer to screen
    ; A1 = start of jump table
    ; D2 = number of scan lines in char
    move.l  monocolmask(pc),d1
    move.w  V_CUR_X(a2),d0
    lsr.w   #1,d0               ; # of words must be even!!!!

.clear:
    move.l  V_CUR_AD(a2),a0
    lea     .endjump(pc),a1
    move.w  BYTES_LIN(a2),d2
    lsr.w   #1,d2
    suba.w  d2,a1

    move.w  V_CEL_HT(a2),d2
    subq.w  #1,d2
    
.floop:
    jmp     0(a1,d0.w)
.jump:
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
    move.l  d1,(a0)+
.endjump:
    adda.w  d0,a0           ; skip start of next scan line
    adda.w  d0,a0
    dbf     d2,.floop

    bra     .exit

.last:
    clr.w   d0                  ; d4 was 79, so draw one more space
    bsr     outCON2

.exit:
    move.l  a3,V_CUR_AD(a2)     ; restore cursor position
    move.w  d3,V_CUR_X(a2)      ; restore x
    tst.b   d5
    beq.s   .x

    bset    #3,V_CURSTAT(a2)

.x:
    movem.l (sp)+,d3-d6/a3    
    rts

;
; ESC o - clear from start of  line, cursor position unchanged
;

vt_cbol:
    setmode outCON
    bsr     set_plot

    movem.l d3/a3,-(sp)

    move.l  V_CUR_AD(a2),a3
    move.w  V_CUR_X(a2),d3     ; get x
    beq.s   .x                  ; if 0, nothing to delete

    clr.w   V_CUR_X(a2)
    move.l  plineTEXT(pc),V_CUR_AD(a2)

.loop:
    clr.w   d0
    bsr     outCON2

    cmp.w   V_CUR_X(a2),d3
    bne     .loop

    move.l  a3,V_CUR_AD(a2)
    move.w  d3,V_CUR_X(a2)
.x:
    movem.l (sp)+,d3/a3
    rts


;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

;
;
;   TRAP #2 VDI handler
;
;
;
;

    ; GDOS has intercepted TRAP #2 (booooo). It has JSRed to this point.
    ; Right now the supervisor stack contains the interrupt frame,
    ; and return address to GDOS when we're finished.
    ; We have to fake another interupt frame, jump into our trap #2 handler,
    ; and when it RTEs back to us, check D0. If it is 0, we successfully
    ; executed a command, otherwise it should be $73 and we have to call TOS.
    
GDOShook:
    pea     .1              ; fake an interrupt frame
    move.w  SR,-(sp)
    bra     doTrap2

.1:
    ; Quick ST's routines will RTE to this point.
    ; Check D0 for return status
    cmpi.w   #115,d0
    bne.s    .2

    ; go through ROM
    move.l   vdiROMvec(pc),-(sp)

    ; go back to NFG GDOS
.2:
    rts

AUTOhook:
    cmpi.w   #$FFFF,D0
    bne.s    oldTrap2
    move.l   #GDOShook,D0
    rte

myTrap2:
    cmpi.w   #115,D0
    bne.s    AUTOhook

doTrap2:
    move.l   a0,d0          ; save A0
    move.l   d1,a0          ; get VDI parameter block
    move.l   (a0)+,a0       ; get contrl array
    move.w   (a0),a0        ; get command # from contrl[0]
    exg      a0,d0          ; restore A0, put command in D0
    add.w    d0,d0
    add.w    d0,d0
    move.l   mpTrap2(PC,d0.w),-(sp)
    moveq    #115,d0
    rts

; jump here to abort custom VDI and continue with ROM VDI

goTrap2:
    move.l  (sp)+,a2
    move.l  (sp)+,a0
    moveq   #115,d0

oldTrap2:
    jmp     12345678

; jump here at the end of a custom AES/VDI routine

noTrap2:
    move.l  (sp)+,a2
    move.l  (sp)+,a0
    moveq   #0,d0
    rte

mpTrap2:
    dcb.l   144,0           ; jump table for trap #2 commands

    ; common code for all custom VDI routines that use workstation handles
    .macro  comTrap2
    move.l  a0,-(sp)
    move.l  a2,-(sp)

    lea     qcontrl(pc),a2
    move.l  d1,a0           ; get VDI parameter block
    move.l  (a0)+,(a2)+     ; contrl
    move.l  (a0)+,(a2)+     ; intin
    move.l  (a0)+,(a2)+     ; ptsin
    move.l  (a0)+,(a2)+     ; intout
    move.l  (a0),(a2)       ; ptsout

    move.l  qcontrl(pc),a0
    move.w  12(a0),d0       ; VDI handle

    cmpi.w  #10,d0
    bcc     goTrap2         ; make sure handle number is less than 10

    tst.w   fGplus
    beq.s   .noGP

    move.l  lGplusBD(pc),a2 ; we have a virtual handle
    move.l  26(a2),a2       ; pointer to handle table
    adda.w  d0,a2
    move.b  (a2),d0         ; get physical handle
    bra.s   .noGDOS

.noGP:
    ; Just in case GDOS is loaded beneath us, map virtual handles.
    ; This doesn't really work! A better way needs to be found.

    cmpi.w  #3,d0
    blt.s   .setno

    tst.w   fGDOS
    beq.s   .noGDOS

    subq.w  #2,d0       ; 3 -> 1, 5 -> 3, 7-> 5, 8 -> 6
    cmpi.w  #1,d0
    beq.s   .noGDOS

    subq.w  #1,d0       ; 5 -> 2, 7 -> 4, 8 -> 5
    bra.s   .noGDOS

.setno:
    clr.w   fGDOS
    
.noGDOS:
    move.w  pLineA(pc),a2   ; lineA pointer
    move.l  CUR_WORK(a2),a0 ; pointer to virtual workstation attributes
.vw:
    cmp.w   $28(a0),d0      ; is current workstation the one we want?
    beq.s   .letsgo

    tst.l   $40(a0)         ; pointer to next workstation structure
    beq     goTrap2         ; can't do if current handle != handle
    move.l  $40(a0),a0      ; get next virtual workstation
    bra     .vw

.letsgo:
    .endm

v_rectfl:
    comTrap2
    ; at this point we're ready to plot, need to check write mode and color
    cmpi.w  #0,$128(a0)     ; check write mode
    bne     goTrap2

    cmpi.w  #0,$1E(a0)      ; check fill color
    beq.s   .vrf1           ; 0 is white

    cmpi.w  #0,$24(a0)      ; check fill interior style
    bne     goTrap2         ; 0 is not filled

.vrf1:
    movem.l d1-d7/a0-a6,-(sp)
    move.l  a0,a3           ; current virtual workstation
    move.l  qcontrl(pc),a6  ; get contrl array
    cmpi.w  #2,2(a6)        ; check contrl[1]
    bne     av_rectfl
    clr.w   4(a6)           ; contrl[2] = 0
    clr.w   8(a6)           ; contrl[4] = 0
    move.l  qptsin(pc),a0   ; get ptsin

    move.w  (a0)+,d1        ; get x1
    bpl.s   .x1
    clr     d1              ; x can't be negative!
.x1:
    move.w  (a0)+,d2        ; get y1
    bpl.s   .y1
    clr     d2              ; y can't be negative!
.y1:
    move.w  (a0)+,d3        ; get x2
    bpl.s   .x2
    clr     d3              ; x can't be negative!
.x2:
    move.w  (a0)+,d4        ; get y2
    bpl.s   .y2
    clr     d4              ; y can't be negative!
.y2:

    ; make sure x2>x1 and y2>y1, swap if necessary
    cmp.w   d1,d3
    bgt.s   .vrf1x
    exg     d1,d3
.vrf1x:
    cmp.w   d2,d4
    bgt.s   .vrf1y
    exg     d2,d4
.vrf1y:

    ; now check workstation clipping
    move.w  $02(a3),d0      ; workstation clip flag
    beq.s   .clip

    move.w  $130(a3),D0     ; get VDI workstation min Y clipping
    bmi     .1

    cmp.w   d0,d2
    bcc.s   .1
    move.w  d0,d2
    cmp.w   d2,d4
    bcs     xv_rectfl       ; clipped right out of existence
.1:
    move.w  $132(a3),d0     ; get ymaxcl
    cmp.w   d0,d4
    blt.s   .2

    move.w  d0,d4     ; max y clip
    cmp.w   d2,d4
    bcs     xv_rectfl       ; clipped right out of existence
.2:
    move.w  $12C(a3),D0     ; get VDI workstation min X clipping
    bmi     .3

    cmp.w   d0,d1
    bgt.s   .3
    move.w  d0,d1
    cmp.w   d1,d3
    bcs     xv_rectfl       ; clipped right out of existence
.3:
    move.w  $12E(a3),d0     ; get xmaxcl
    cmp.w   d0,d3
    blt.s   .clip

    move.w  d0,d3           ; max x clip
    cmp.w   d1,d3
    bcs     xv_rectfl       ; clipped right out of existence

.clip:    
    move.w  d1,d5
    move.w  d3,d6
    and.w   #31,d5
    and.w   #31,d6
    add.w   d5,d5
    add.w   d5,d5
    add.w   d6,d6
    add.w   d6,d6
    lea     mapxtolmask(pc),a0
    move.l  0(a0,d5.w),d5
    move.l  4(a0,d6.w),d6
    not.l   d6

    move.w  d2,d0
    mulu    BYTES_LIN(a2),d0    ; vertical offset into screen
    move.l  $44e,a0         ; start of screen
    adda.l  d0,a0           ; add offset into screen

    move.w  d1,d0
    lsr.w   #3,d0
    and.w   #$fffc,d0       ; longword boundary
    adda.w  d0,a0           ; first screen byte
    sub.w   d2,d4           ; number of scan lines + 1

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; register usage:
;       a0 - pointer to scan line
;       a1 - pointer to actual screen byte
;       d0 - horizontal long counter
;       d3 - # of horizontal longwords - 1
;       d4 - vertical loop counter
;       d5 - left clipping mask
;       d6 - right clipping mask
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    lsr.w   #5,d1
    lsr.w   #5,d3
    sub.w   d1,d3           ; number of long words

    bne.s   .vrf2

    ; special case, both points in same longword
    or.l    d5,d6
.samelw:
    and.l   d6,(a0)
    adda.w  BYTES_LIN(a2),a0
    dbf     d4,.samelw
    bra     xv_rectfl
    
.vrf2:
    subq.w  #1,d3
.vrf22:
    move.w  d3,d0
    movea.l a0,a1

    clr.l   d1
    and.l   d5,(a1)+            ; left clipping
    bra     .vrf4

.vrf3:
    move.l  d1,(a1)+
.vrf4:
    dbf     d0,.vrf3
    and.l   d6,(a1)             ; right clipping
    adda.w  BYTES_LIN(a2),a0
    dbf     d4,.vrf22

xv_rectfl:
    movem.l (sp)+,d1-d7/a0-a6
    bra     noTrap2

av_rectfl:
    movem.l (sp)+,d1-d7/a0-a6
    bra     goTrap2

mapxtolmask:
    dc.l    $00000000
    dc.l    $80000000
    dc.l    $c0000000
    dc.l    $e0000000
    dc.l    $f0000000
    dc.l    $f8000000
    dc.l    $fc000000
    dc.l    $fe000000
    dc.l    $ff000000
    dc.l    $ff800000
    dc.l    $ffc00000
    dc.l    $ffe00000
    dc.l    $fff00000
    dc.l    $fff80000
    dc.l    $fffc0000
    dc.l    $fffe0000
    dc.l    $ffff0000
    dc.l    $ffff8000
    dc.l    $ffffc000
    dc.l    $ffffe000
    dc.l    $fffff000
    dc.l    $fffff800
    dc.l    $fffffc00
    dc.l    $fffffe00
    dc.l    $ffffff00
    dc.l    $ffffff80
    dc.l    $ffffffc0
    dc.l    $ffffffe0
    dc.l    $fffffff0
    dc.l    $fffffff8
    dc.l    $fffffffc
    dc.l    $fffffffe
    dc.l    $ffffffff

v_ESC:
    comTrap2
    move.w  10(a0),d0
    subi.w  #2,d0           ; only allow VDI ESC 2 thru 14
    bcs     goTrap2
    cmpi.w  #13,d0
    bcc     goTrap2

    add.w   d0,d0
    add.w   d0,d0
    move.l  vdiESCjmp(PC,d0.w),a0
    jmp     (a0)

vdiESCjmp:
    dc.l    vdi_exit_cur
    dc.l    vdi_enter_cur
    dc.l    vdi_curs_up
    dc.l    vdi_curs_down
    dc.l    vdi_curs_right
    dc.l    vdi_curs_left
    dc.l    vdi_curs_home
    dc.l    vdi_ceos
    dc.l    vdi_eeol
    dc.l    vdi_curs_pos
    dc.l    vdi_bconout
    dc.l    vdi_inv_on
    dc.l    vdi_inv_off

; VDI ESC 2

vdi_exit_cur:
    movem.l d1-d3/a1,-(sp)
    bsr     vt_off
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 3

vdi_enter_cur:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)   ; vt_on is one based
    bsr     vt_cls
    bsr     vt_on
    subq.w  #1,V_HID_CNT(a2)   ; vt_on is one based
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 4

vdi_curs_up:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_up
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 5

vdi_curs_down:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_down
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 6

vdi_curs_right:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_right
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 7

vdi_curs_left:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_left
    subq.w  #1,V_HID_CNT(a2)
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 8

vdi_curs_home:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_home
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 9

vdi_ceos:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_ceos
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 10

vdi_eeol:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_eeol
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 11

vdi_curs_pos: 
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     clear_cursor
    move.l  qintin(pc),a0
    move.w  (a0)+,d2        ; get Y
    move.w  (a0),d1         ; get X
    subq.w  #1,d1           ; VDI is one based, VT52 is 0 based
    subq.w  #1,d2
    bsr     new_curs_pos
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 12

vdi_bconout:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    move.l  qintin(pc),a0  ; get intin array
    move.w  (a0),d0         ; get character
    ext.w   d0
    bsr     outCON
    subq.w  #1,V_HID_CNT(a2)
    bne.s   .x
    bsr     set_cursor
    bset    #2,V_CURSTAT(a2)
.x:
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 13

vdi_inv_on:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_set_inv
    subq.w  #1,V_HID_CNT(a2)
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2

; VDI ESC 14

vdi_inv_off:
    movem.l d1-d3/a1,-(sp)
    addq.w  #1,V_HID_CNT(a2)
    bsr     vt_clear_inv
    subq.w  #1,V_HID_CNT(a2)
    movem.l (sp)+,d1-d3/a1
    bra     noTrap2


vst_alignment:
    comTrap2
    bra     goTrap2

;
;   Custom VDI routines
;
;

;
; All the useless set this and set that commands.
; These are implemented only because they are called
; so often that there is a lot of overhead associated
; with them. They coulda made workstation pointers
; visible to a VDI app. No, that would be too (*&^*( easy.
;

vst_color:
    comTrap2
    move.l  qcontrl(pc),a2
    move.w  #1,8(a2)
    move.l  qintin(pc),a2
    move.w  (a2),d0
    beq.s   .ok
    moveq   #15,d0
.ok:
    move.w  d0,$A2(a0)          ; $A2 = text color
    move.l  qintout(pc),a2
    andi.w  #1,d0
    move.w  d0,(a2)
    bra     noTrap2

vsf_color:
    comTrap2
    move.l  qcontrl(pc),a2
    move.w  #1,8(a2)
    move.l  qintin(pc),a2
    move.w  (a2),d0
    beq.s   .ok
    moveq   #15,d0
.ok:
    move.w  d0,$1E(a0)          ; $1E = fill pattern color
    move.l  qintout(pc),a2
    andi.w  #1,d0
    move.w  d0,(a2)
    bra     noTrap2

vsl_color:
    comTrap2
    move.l  qcontrl(pc),a2
    move.w  #1,8(a2)
    move.l  qintin(pc),a2
    move.w  (a2),d0
    beq.s   .ok
    moveq   #15,d0
.ok:
    move.w  d0,$2C(a0)          ; $2C = polyline color
    move.l  qintout(pc),a2
    andi.w  #1,d0
    move.w  d0,(a2)
    bra     noTrap2

vswr_mode:
    comTrap2
    move.l  qcontrl(pc),a2
    move.w  #1,8(a2)
    move.l  qintin(pc),a2
    move.w  (a2),d0
    subq.w  #1,d0
    cmpi.w  #4,d0
    bcs.s   .ok
    moveq   #0,d0
.ok:
    move.w  d0,$128(a0)     ; $128 = write mode in workstation
    addq.w  #1,d0
    move.l  qintout(pc),a2
    move.w  d0,(a2)
    bra     noTrap2


vsl_type:
    comTrap2
    move.l  qcontrl(pc),a2
    move.w  #1,8(a2)
    move.l  qintin(pc),a2
    move.w  (a2),d0
    subq.w  #1,d0
    cmpi.w  #7,d0
    bcs.s   .ok
    moveq   #0,d0
.ok:
    move.w  d0,$30(a0)      ; $30 = polyline line type in workstation
    addq.w  #1,d0
    move.l  qintout(pc),a2
    move.w  d0,(a2)
    bra     noTrap2

    ; udsty - set up line pattern variable
vsl_udsty:
    comTrap2
    move.l  qintin(pc),a2   ; get intin
    move.w  (a2),$a4(a0)    ; $A4 = user defined pattern
    bra     noTrap2

v_pline:
    comTrap2
    cmpi.w  #1,$32(a0)      ; check line thickness
    bne     goTrap2         ; != 1, can't do

    movem.l d1-d7/a0-a6,-(sp)
    move.l  a0,a3           ; current virtual workstation
    move.l  qcontrl(pc),a6  ; get contrl array
    clr.w   4(a6)           ; contrl[2] = 0
    clr.w   8(a6)           ; contrl[4] = 0
    move.w  2(a6),d7        ; get count of points in contrl[1]
    subq.w  #2,d7
    bmi.s   av_pline        ; make sure > 1
    move.l  qptsin(pc),a6   ; get ptsin

    move.w  d7,d0
    move.l  a6,a0

    move.w  (a0)+,d3        ; get initial x
    move.w  (a0)+,d4        ; get initial y

.loop:
    move.w  (a0)+,d1        ; get x
    move.w  (a0)+,d2        ; get y
    cmp.w   d1,d3
    beq.s   .next
    cmp.w   d2,d4
    bne.s   av_pline        ; a diagonal line. abort! abort!
.next:
    move.w  d1,d3
    move.w  d2,d4
    dbf     d0,.loop

    move.w  $30(a3),d0
    cmpi.w  #6,d0           ; check line type
    bne.s   .normal

    move.w  $a4(a3),wLineStyle
    move.w  $a4(a3),lLineStyle
    bra     v_1pline        ; do mono v_pline

.normal:
    add.w   d0,d0
    move.w  .mpLineStyle(pc,d0.w),d0
    move.w  d0,wLineStyle
    move.w  d0,lLineStyle
    bra     v_1pline        ; do mono v_pline

.mpLineStyle:
    dc.w    $FFFF
    dc.w    $FFF0
    dc.w    $E0E0
    dc.w    $FE38
    dc.w    $FF00
    dc.w    $F198

xv_pline:
    movem.l (sp)+,d1-d7/a0-a6
    bra     noTrap2

av_pline:
    movem.l (sp)+,d1-d7/a0-a6
    bra     goTrap2

v_1pline:
    ; at this point, we have all straight hor. or vert. lines, 1 pixel wide
    ; d7.w  = count of points - 2 = number of lines - 1
    ; d6.l  = line style
    ; d5.l  = bit mask
    ; d4.w  = pixel counter for each line
    ; d1, d2, d3, d4 = x1, y1, x2, y2 during addree calc
    ;
    ; a6 = ptsin
    ; a5    = pointer to bit mask table
    ; a4    = pointer to scan line offset table
    ; a1    = direction vector (for vertical)
    ; a0    = pointer to screen byte (32 bit aligned)

    clr.w   d1
    clr.w   d2
    move.w  V_REZ_HZ(a2),d3 ; screen width
    move.w  V_REZ_VT(a2),d4 ; screen height

    move.w  $02(a3),d0      ; workstation clip flag
    beq.s   .clip

    move.w  $130(a3),D0     ; get VDI workstation min Y clipping
    bmi     .1

    move.w  d0,d2
.1:
    move.w  $132(a3),d0     ; get ymaxcl
    cmp.w   d0,d4
    blt     .2

    move.w  d0,d4     ; max y clip
.2:
    move.w  $12C(a3),D0     ; get VDI workstation min X clipping
    bmi     .3

    move.w  d0,d1
.3:
    move.w  $12E(a3),d0     ; get xmaxcl
    cmp.w   d0,d3
    blt     .clip

    move.w  d0,d3           ; max x clip

.clip:
    move.w  d1,xmincl
    move.w  d2,ymincl
    move.w  d3,xmaxcl
    move.w  d4,ymaxcl

    move.w  BYTES_LIN(a2),a1
    move.l  lLineStyle(pc),d6       ; line style (32 bits)

vpl_loop:
    move.w  (a6)+,d1        ; get x
    move.w  (a6)+,d2        ; get y

    move.w  (a6),d3         ; get next x
    move.w  2(a6),d4        ; get next y

    ; can't calculate screen address quite yet, due to possible clipping
    cmp.w   d1,d3           ; x1 != x2?
    bne     hor_lin_m       ; must be a horizontal line then

vert_lin_m:
    ; d1 == d3, check x for clipping
    cmp.w   xmincl(pc),d1
    blt     vpl_next        ; x is too small
    cmp.w   xmaxcl(pc),d1
    bgt     vpl_next           ; x is too large

    ; make sure y1 < y2
    cmp.w   d2,d4
    bgt.s   .ok
    exg     d2,d4

.ok:
    ; now check d2 and d4 for clipping
    cmp.w   ymincl(pc),d2       ; y1 > min clip
    bge.s   .v1

    move.w  ymincl(pc),d2
    cmp.w   ymincl(pc),d4       ; y2 > min clip else abort
    blt     vpl_next

.v1:
    cmp.w   ymaxcl(pc),d4       ; y2 < max clip
    blt.s   .v2

    move.w  ymaxcl(pc),d4
    cmp.w   ymaxcl(pc),d2       ; y1 < max clip else abort
    bgt     vpl_next

.v2:
    ; calculate address of initial point
    move.w  a1,d0
    mulu    d2,d0           ; vertical offset into screen
    move.l  $44e,a0         ; start of screen
    adda.l  d0,a0           ; add offset into screen

    lsr.w   #3,d3
    andi.w  #$FFFE,d3       ; word align
    adda.w  d3,a0           ; first screen word

    andi.w  #$0F,d1
    add.w   d1,d1
    lea     rgwBitMasks(pc),a5
    move.w  0(a5,d1.w),d5

    sub.w   d2,d4           ; calculate y2-y1
    beq     vpl_dot
    bmi     vpl_next

.down:
    move.l  d6,d3           ; line style needs to be rotated

    cmpi.w  #2,$128(a3)     ; XOR mode
    beq     .downx

    tst.w   $2C(a3)         ; check line color
    bne.s   .downt

    not.w   d5
.downw:
    rol.w   d3
    bcc.s   .downw2
    and.w   d5,(a0)         ; plot pixel, white
.downw2:
    adda.w  a1,a0
    subq.w  #1,d4
    bne     .downw
    bra     vpl_next

.downt:
    cmpi.w  #17,d4
    blt.s   .downt1         ; check if the line is long enough
    cmpi.w  #$FFFF,d6
    beq.s   .downtFF
.downt1:
    rol.w   d3
    bcc.s   .downt2
    or.w    d5,(a0)         ; plot pixel, black
.downt2:
    adda.w  a1,a0
    subq.w  #1,d4
    bne     .downt1
    bra     vpl_next

.downtFF:
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0
    or.w    d5,(a0)
    adda.w  a1,a0

    sub.w   #16,d4
    cmpi.w  #16,d4
    bgt     .downtFF

.downtFF2:
    or.w    d5,(a0)
    adda.w  a1,a0
    subq.w  #1,d4
    bgt     .downtFF2
    bra     vpl_next

.downx:
    cmpi.w  #17,d4
    blt.s   .downx1         ; check if the line is long enough
    cmpi.w  #$AAAA,d6       ; to bother checking for special patterns
    beq.s   .downxAA
    cmpi.w  #$5555,d6
    beq.s   .downx55
.downx1:
    rol.w   d3
    bcc.s   .downx2
    eor.w    d5,(a0)         ; plot pixel, XOR
.downx2:
    adda.w  a1,a0
    subq.w  #1,d4
    bne     .downx1
    bra     vpl_next

.downxAA:
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
.downx55:
    adda.w  a1,a0
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    sub.w   #16,d4
    cmpi.w  #16,d4
    bgt     .downxAA

.downxAA2:
    eor.w    d5,(a0)         ; plot pixel, XOR
    adda.w  a1,a0
    adda.w  a1,a0
    subq.w  #2,d4
    bgt     .downxAA2
    bra     vpl_next


hor_lin_m:
    ; d2 == d4, check y for clipping
    cmp.w   ymincl(pc),d2
    blt     vpl_next           ; y is too small
    cmp.w   ymaxcl(pc),d2
    bgt     vpl_next           ; y is too large

    cmp.w   d1,d3
    bgt.s   .ok
    exg     d1,d3

.ok:
    ; now check d1 and d3 for clipping
    cmp.w   xmincl(pc),d1       ; x1 > min clip
    bge     .h1

    move.w  xmincl(pc),d1
    cmp.w   xmincl(pc),d3       ; x2 > min clip else abort
    blt     vpl_next

.h1:
    cmp.w   xmaxcl(pc),d3       ; x2 < max clip
    blt     .h2

    move.w  xmaxcl(pc),d3
    cmp.w   xmaxcl(pc),d1       ; x1 < max clip else abort
    bgt     vpl_next

.h2:
    ; calculate address of initial point
    move.w  a1,d0
    mulu    d2,d0           ; vertical offset into screen
    move.l  $44e,a0         ; start of screen
    adda.l  d0,a0           ; add offset into screen

    move.w  d1,d0
    lsr.w   #3,d0
    andi.w  #$FFFC,d0       ; long word align
    adda.w  d0,a0           ; first screen long word

    move.w  d1,d0           ; left clipping mask
    andi.w  #$1F,d0
    add.w   d0,d0
    add.w   d0,d0
    lea     mapxtolmask(pc),a5
    move.l  0(a5,d0.w),d2
    not.l   d2

    move.w  d3,d0
    andi.w  #$1F,d0         ; right clipping mask
    add.w   d0,d0
    add.w   d0,d0
    move.l  4(a5,d0.w),d4

    lsr.w   #5,d1
    lsr.w   #5,d3
    sub.w   d1,d3           ; number of longwords - 1

    bgt.s   .hor
    beq     .samel
    illegal

    ; special case: xmin and xmax are in the same longword
.samel:
    and.l   d2,d4
    and.l   d6,d4

    cmpi.w  #2,$128(a3)     ; XOR mode
    beq.s   .samex

    tst.w   $2C(a3)         ; check line color
    bne.s   .samet

    not.l   d4
    and.l   d4,(a0)
    bra     vpl_next

.samet:
    or.l    d4,(a0)
    bra     vpl_next

.samex:
    eor.l   d4,(a0)
    bra     vpl_next

.hor:
    cmpi.w  #2,$128(a3)     ; XOR mode
    beq.s   .horx

    tst.w   $2C(a3)         ; check line color
    bne.s   .hort

    move.l  d6,d0
    move.l  d0,d1
    not.l   d0              ; negative pixel mask
    and.l   d2,d1
    not.l   d1

    and.l   d1,(a0)+
    subq.w  #2,d3
    bmi.s   .horwr
.horw:
    and.l   d0,(a0)+
    dbf     d3,.horw
.horwr:
    and.l   d6,d4           ; do right side
    not.l   d4
    and.l   d4,(a0)+
    bra     vpl_next

.hort:
    move.l  d6,d0
    move.l  d0,d1
    and.l   d2,d1           ; do left side
    or.l    d1,(a0)+
    subq.w  #2,d3
    bmi.s   .hortr
.hort1:
    or.l    d6,(a0)+
    dbf     d3,.hort1
.hortr:
    and.l   d6,d4           ; do right side
    or.l    d4,(a0)+
    bra     vpl_next

.horx:
    move.l  d6,d0
    move.l  d0,d1
    and.l   d2,d1           ; do left side
    eor.l   d1,(a0)+
    subq.w  #2,d3
    bmi.s   .horxr
    moveq   #8,d0
    cmp.w   d0,d3
    blt.s   .horx1
.horxq:
    eor.l   d6,(a0)+
    eor.l   d6,(a0)+
    eor.l   d6,(a0)+
    eor.l   d6,(a0)+
    eor.l   d6,(a0)+
    eor.l   d6,(a0)+
    eor.l   d6,(a0)+
    eor.l   d6,(a0)+
    sub.w   d0,d3
    cmp.w   d0,d3
    bcc     .horxq
.horx1:
    eor.l   d6,(a0)+
    dbf     d3,.horx1
.horxr:
    and.l   d6,d4           ; do right side
    eor.l   d4,(a0)+
    bra     vpl_next


vpl_dot:
    ; we don't do single pixel lines!

vpl_next:
    dbf     d7,vpl_loop

    bra     xv_pline

rgwBitMasks:
    dc.w    $8000
    dc.w    $4000
    dc.w    $2000
    dc.w    $1000
    dc.w    $0800
    dc.w    $0400
    dc.w    $0200
    dc.w    $0100
    dc.w    $0080
    dc.w    $0040
    dc.w    $0020
    dc.w    $0010
    dc.w    $0008
    dc.w    $0004
    dc.w    $0002
    dc.w    $0001

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   VDI v_gtext routines
;
; register usage:
;
;   A0 -
;   A1 - pointer to v_gtext string
;   A2 - pointer to line A structure / screen pointer
;   A3 - pointer to current workstation attributes
;   A4 - pointer to font header
;   A5 - pointer to font data
;   A6 - pointer to plotting routine
;
;
;   D0 -
;   D1 -
;   D2 -
;   D3 - min X clipping (lo word) / clip mask (hi word)
;   D4 - max X clipping / clip mask
;   D5 - count of chars to print
;   D6 - X co-ordinate of text
;   D7 - Y co-ordinate of text / screen offset
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

left_clip: dc.w 0
right_clip: dc.w 0

rgbClipMask:
    dc.b    $00, $01, $03, $07, $0F, $1F, $3F, $7F

v_gtext:
    comTrap2

    movem.l d1-d7/a0-a6,-(sp)
    move.l  a0,a3           ; current virtual workstation
    lea     qcontrl(pc),a0
    move.l  (a0)+,a6        ; get contrl array
    clr.w   4(a6)           ; contrl[2] = 0
    clr.w   8(a6)           ; contrl[4] = 0
    move.w  6(a6),d5        ; get count of chars in contrl[3]
    move.l  (a0)+,a1        ; get intin (array of 16 bit text)
    move.l  (a0)+,a6        ; get ptsin
    move.w  (a6)+,d6        ; x co-ordinate
    move.w  (a6),d7         ; y co-ordinate

; for now, call the old VDI routines for things we can't do

    move.w  d6,d0
    andi.w  #$07,d0
    beq     .0              ; it's byte aligned text

    move.w  d0,wShiftBits

.0:
    tst.w   $1A(a3)         ; check style, we don't do bold, italic, etc...
    bne     .oldVDI

    move.l  4(a3),a4        ; pointer to current font header

    cmpi.w  #8,$34(a4)      ; check font width
    bne     .oldVDI         ; probably 6x6 font

    move.w  $42(a4),d0
    btst    #3,d0
    beq     .oldVDI         ; not a monospace font

.00:
    ; at this point, the workstation handle is OK, the font is OK, and
    ; the text is not bold or italic, etc...
    ; Let's rock 'n roll!

    cmpi.w  #5,$126(a3)     ; check vertical font alignment == top line
    beq.s   .vert
    sub.w   $28(a4),D7      ; font alignment
    cmpi.w  #3,$126(a3)     ; check vertical font alignment == bottom
    bne.s   .vert
    sub.w   $30(a4),D7      ; font alignment
    ; missing a few alignment checks are we?

.vert:
    cmp.w   V_REZ_VT(a2),d7
    bcc     .oldVDI         ; exceeds bottom of screen, we're done!

    move.w  V_REZ_HZ(a2),D4 ; get screen size as default clipping
    clr.w   d3
    clr.l   left_clip

    tst.w   $02(a3)         ; workstation clip flag
    beq.s   .13

    cmp.w   $130(a3),D7     ; workstation Y clipping
    bcs     .oldVDI
    move.w  d7,d0
    add.w   $52(a4),d0      ; add character height
    subq.w  #1,d0           ; bottom scan line of text
    cmp.w   $132(a3),d0     ; compare to ymaxcl
    bhi     .oldVDI

    move.w  $12E(a3),D3     ; workstation X clipping
    addq.w  #1,d3           ; treat as a mac
    cmp.w   d4,d3
    bcc.s   .1
    move.w  d3,d4           ; smaller, so use it

.1:
    move.b  d4,d0
    andi.w  #$07,d0
    move.w  d0,right_clip
    addq.w  #7,d4
    andi.w  #$FFF8,d4       ; round up clipping to byte boundary

    move.w  $12C(a3),d3     ; workstation minimum X clip

    cmp.w   d3,d6           ; is xmincl < x
    bcc.s   .13             ; yep, no need to worry about min clip

    move.b  d3,d0
    andi.w  #$07,d0
    move.w  d0,left_clip
    andi.w  #$FFF8,d3         ; round down clipping to byte boundary

    ; left edge clipping
.12:
    cmp.w   d3,d6           ; is xmincl < x
    bcc.s   .13             ; yep

    ; clip off the first few characters of the string
    addq.l  #8,d6           ; increment x
    move.w  (a1)+,d0        ; go to next string char
    beq     .xv_gtext       ; it was 0, end of string, done!
    subq.l  #1,d5           ; decrement cch
    beq     .xv_gtext
    bra     .12

.13:
    mulu    BYTES_LIN(a2),d7    ; vertical offset into screen

    move.w  d4,d2
    sub.w   d6,d2           ; subtact x from xmaxcl
    bcs     .xv_gtext       ; if x > xmaxcl, then nothing to print

    lsr.w   #3,D2           ; convert to number of chars
    cmp.w   d5,d2           ; compare to string length
    bcc.s   .norclip        ; if less, use clipped length
    move.w  d2,d5
    bra     .2

.norclip:
    clr.w   right_clip

.2:
    subq.w  #1,d5           ; max number of chars to print - 1
    bmi     .xv_gtext

    move.w  right_clip(pc),d0
    sne     d0
    ext.w   d0              ; = -1 if right_clip
    add.w   d0,d5           ; decrement cch is right_clip

.1begin:
    move.w  d6,d1
    lsr.w   #3,D1           ; convert pixels to bytes
    ext.l   d1
    add.l   d1,d7           ; offset into screen

    andi.w  #7,d6
    bne.s   .1sgen

    move.l  $4C(a4),a5      ; font data
    lea     vgt1gen(pc),a6
    bra     .1start

.1sgen:
    move.l  $4C(a4),a5      ; font data
    lea     vgt1sgen(pc),a6
    bra     .1start

.1start:
    move.w  $A2(a3),d0      ; text color
    andi.w  #1,d0
    add.w   d0,d0
    add.w   d0,d0
    adda.w  d0,a6

    move.w  $128(a3),d0     ; write mode
    andi.w  #3,d0
    add.w   d0,d0
    add.w   d0,d0
    add.w   d0,d0
    adda.w  d0,a6

    move.l  (a6),a6         ; get address of plotting routine

.start:
    move.w  BYTES_LIN(a2),d1
    move.w  V_FNT_WD(a2),d2

    move.l  $44e,A2         ; start of screen
    adda.l  d7,A2           ; add offset into screen

    tst.w   d5              ; if cch == -1
    bmi     .rclip

.s2:
    move.w  left_clip(pc),d0
    beq.s   .loop

    add.w   d0,d0
    move.w  .LeftMask(PC,d0.w),d0
    move.w  d0,d4       ; mask for font
    not     d0
    move.w  d0,d6       ; mask for screen data

    ; first character is clipped on the left side
    move.w  (a1)+,D0
    beq.s   .xv_gtext
    cmpi.w  #$100,D0
    bcc.s   .1next
    ext.w   d0

    move.l  a5,A0       ; get pointer to font
    move.l  a2,a3       ; save cursor position
    jsr     -4(a6)      ; go plot character
    move.l  a3,a2       ; restore original cursor position
    bra     .1next

.LeftMask:
    dc.w    $0000, $7F7F, $3F3F, $1F1F, $0F0F, $0707, $0303, $0101

.loop:
    move.w  (a1)+,D0
    beq.s   .xv_gtext
    ext.w   d0

    move.l  a5,A0       ; get pointer to font

    move.l  a2,a3       ; save cursor position
    jsr     (a6)        ; go plot character
    move.l  a3,a2       ; restore original cursor position
.1next:
    addq.l  #1,a2       ; increment cursor position
    dbf     d5,.loop    ; next character

    move.w  right_clip(pc),d0
    beq.s   .xv_gtext

.rclip:
    add.w   d0,d0
    move.w  .RightMask(PC,d0.w),d0
    move.w  d0,d4       ; mask for font
    not     d0
    move.w  d0,d6       ; mask for screen data

    ; last character is clipped on the right side
    move.w  (a1),D0
    beq.s   .xv_gtext
    ext.w   d0

    move.l  a5,A0       ; get pointer to font
    jsr     -4(a6)      ; go plot character

.xv_gtext:
    movem.l (sp)+,d1-d7/a0-a6
    bra     noTrap2

.RightMask:
    dc.w    $0000, $8080, $C0C0, $E0E0, $F0F0, $F8F8, $FCFC, $FEFE

.oldVDI:
    movem.l (sp)+,d1-d7/a0-a6
    bra     goTrap2



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; v_gtext plotting routines
;
; naming convention:
;
;   vgt[1|2|4][0-9,A-F][r|t|x|i][88|816|gen][n|c]
;
;       1, 2, 4 - number of bit planes
;
;       0-9,A-F - text color
;
;       r - replace
;       t - transparent
;       x - XOR
;       i - inverse
;
;       88, 816, gen - font size or
;
;       n - normal (byte aligned)
;       c - clipped (byte aligned)
;
; Note: there are 1192 possible combinations
;       but only 48 are implemented in mono
;       and 64 in medium rez, none in lo rez
;
; Register usage:
;
;   A0 -
;   A1 - pointer to v_gtext string
;   A2 - pointer to line A structure / screen pointer
;   A3 - pointer to current workstation attributes
;   A4 - pointer to font header
;   A5 - pointer to font data
;   A6 - pointer to plotting routine
;
;
;   D0 -
;   D1 -
;   D2 -
;   D3 - min X clipping (lo word) / clip mask (hi word)
;   D4 - max X clipping / clip mask
;   D5 - count of chars to print
;   D6 - X co-ordinate of text
;   D7 - Y co-ordinate of text / screen offset
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; monochrome jump table

vgt1gen:
    dc.l    vgt10rgenn
    dc.l    vgt11rgenn
    dc.l    vgt10tgenn
    dc.l    vgt11tgenn
    dc.l    vgt11xgenn      ; XOR mode ignores color
    dc.l    vgt11xgenn
    dc.l    vgt10igenn
    dc.l    vgt11igenn
    
vgt1sgen:
    dc.l    vgt1s0rgenn
    dc.l    vgt1s1rgenn
    dc.l    vgt1s0tgenn
    dc.l    vgt1s1tgenn
    dc.l    vgt1s1xgenn
    dc.l    vgt1s1xgenn
    dc.l    vgt1s0igenn
    dc.l    vgt1s1igenn



; v_gtext monochrome replace, white

    bra     vgt10rgenc

vgt10rgenn:
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    clr.b   (a2)
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

vgt10rgenc:
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    and.b   d6,(a2)
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

; v_gtext monochrome replace, black

    bra     vgt11rgenc

vgt11rgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines

    not.w   d3
    andi.w  #31,d3      ; limit to 31 scan lines
    add.w   d3,d3
    move.w  d3,d0
    add.w   d3,d3
    add.w   d0,d3       ; multiply by 6
    jmp     .here(PC,d3.w)

.here:
    .rept   30
    move.b  (a0),(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    .endr
    move.b  (a0),(a2)

    rts

vgt11rgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    move.b  (a2),d7
    and.w   d4,d0
    and.w   d6,d7
    or.b    d7,d0
    move.b  d0,(a2)
    
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts


; v_gtext monochrome transparent, white

    bra     vgt10tgenc

vgt10tgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    not     d0
    and.b   d0,(a2)

    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

vgt10tgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    and     d4,d0
    not     d0
    and.b   d0,(a2)

    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

; v_gtext monochrome transparent, black

    bra     vgt11tgenc

vgt11tgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines

    not.w   d3
    andi.w  #31,d3      ; limit to 31 scan lines
    lsl.w   #3,d3
    jmp     .here(PC,d3.w)

.here:
    .rept   30
    move.b  (a0),d0
    or.b    d0,(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    .endr
    move.b  (a0),d0
    or.b    d0,(a2)

    rts

vgt11tgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    and.w   d4,d0
    or.b    d0,(a2)
    
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts


; XOR ignores color
; v_gtext monochrome XOR

    bra     vgt11xgenc

vgt11xgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    eor.b    d0,(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

vgt11xgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    and.w   d4,d0
    eor.b    d0,(a2)
    
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

; v_gtext monochrome inverse transparent, white

    bra     vgt10igenc

vgt10igenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    and.b   d0,(a2)

    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

vgt10igenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    and     d4,d0
    and.b   d0,(a2)

    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

; v_gtext monochrome inverse transparent, black

    bra     vgt11tgenc

vgt11igenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    not     d0
    or.b    d0,(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts

vgt11igenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d3  ; number of scan lines
    subq.w  #1,d3

.loop:
    move.b  (a0),d0
    not     d0
    and.w   d4,d0
    or.b    d0,(a2)
    
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d3,.loop
    rts


; v_gtext shifted monochrome replace, white

    bra     vgt1s0rgenc

vgt1s0rgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    clr.b   d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

vgt1s0rgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

    not     d4          ; can't use D6 as inverse mask

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    and.b   d4,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    not     d4

    rts

    bra     vgt1s0tgenc

vgt1s0tgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    not     d7
    and.b   d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

vgt1s0tgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    and     d4,d7
    not     d7
    and.b   d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

    bra     vgt1s0xgenc

vgt1s0xgenn:
    rts

vgt1s0xgenc:
    rts

    bra     vgt1s0igenc

vgt1s0igenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    and.b   d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

vgt1s0igenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    and.w   d4,d7
    and.b   d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts


; v_gtext shifted monochrome replace, black

    bra     vgt1s1rgenc

vgt1s1rgenn: 
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

vgt1s1rgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

    bra     vgt1s1tgenc

vgt1s1tgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    or.b    (a0),d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

vgt1s1tgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    and.w   d4,d7
    or.b    d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

    bra     vgt1s1xgenc

vgt1s1xgenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    eor.b   d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

vgt1s1xgenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    and.w   d4,d7
    eor.b   d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

    bra     vgt1s1igenc

vgt1s1igenn:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    not     d7
    or.b    d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

vgt1s1igenc:
    andi.w  #$00FF,d0   ; ROM font is not sign extended
    adda.w  d0,a0       ; get address of char data in font
    move.w  $52(a4),d6  ; number of scan lines
    subq.w  #1,d6
    move.w  wShiftBits(pc),d3

.loop:
    movep.w 1(a2),d0    ; move next byte
    swap    d0          ; into bits 31-24
    move.b  (a2),d0     ; move current byte into bits 7-0
    rol.l   d3,d0
    move.b  (a0),d7
    and.w   d4,d7
    not     d7
    or.b    d7,d0
    ror.l   d3,d0
    move.b  d0,(a2)
    swap    d0
    movep.w d0,1(a2)
    adda.w  d2,a0       ; next scan line in font
    adda.w  d1,a2       ; next scan line on screen
    dbf     d6,.loop

    rts

; THAT'S ALL FOLKS!!!!!!
