/* AUTODATE - Auto Boot Date & Time Setter for the Atari ST

         Copyright (c) 1986 Vanya J. Cooper, Tucson, Arizona

                    >> OK TO COPY, NOT TO SELL <<

Permission is granted to copy, reproduce, or otherwise distribute this
program in whole or in part for any purpose provided (1) this copyright
notice is included and (2) any fee charged does not exceed the cost of
the media and reasonable handling or connect time.


Purpose:

This program checks the intelligent keyboard clock and prompts you for
the date and time if it hasn't been set, or displays it if it looks
reasonable.  The idea is that you are reminded to set the clock after
power-up, but don't need to after a reset.  In fact you don't even get
to after a reset.  When some nasty little program (usually mine) hangs
up, I find it very annoying to have to answer questions about whether I
want to set the date and time.

It accepts all the common date/time formats I could think of, including
abreviated month names, am/pm and 24 hour time, so pick your favorite.

I wrote this program, partly to explore the C language (my first real
start-from-scratch program), so there are a few more structures, pointers,
and indirect references than there really need to be, but so it goes.

Installation:

To use place AUTODATE.PRG in the folder \AUTO\ on your bootup disk(s).

Rebuilding:

The program was written for the Alycon C in the developers package.
The assembler startup module is included at the end of the C code as
a comment.  You can use ACSTART instead, but your program will be a
little bigger.  Link QSTART.O (or ACSTART), AUTODATE.O, and OSBIND.O.
It does not need GEMLIB, VDIBIND, or AESBIND.

Comments, bugs, suggestions, flattery, or otherwise can be send to

	Vanya J. Cooper
	Compuserve ID: 71545,611

If you *REALLY* like this program, ask about contributing to my hard
disk fund.

end comment */

/************************************************************************
 ************************************************************************
 ***                                                                  ***
 ***                      A U T O D A T E                             ***
 ***                                                                  ***
 ************************************************************************
 ************************************************************************/

/* Things we need */
#include "osbind.h"

/* Define some parameters */
#define TRUE 1
#define FALSE 0
#define MAXLINE 20
#define THIS_YEAR 1986

/* Bios chararacter i/o calls */
#define pr_char(c) Cconout(c)
#define pr_strg(c) Cconws(c)
#define c_necin() (Cnecin() & 0x00ff)

/* Video functions */
#define clr_scr() pr_strg("\033E")
#define rev_vid() pr_strg("\033p")
#define nor_vid() pr_strg("\033q")
#define era_eol() pr_strg("\033K")

/* Some ascii chars */
#define CH_BEL '\007'
#define CH_ESC '\033'
#define CH_DEL '\177'

/* DOS date/time structures */
struct dos_date {
    unsigned day : 5;
    unsigned month : 4;
    unsigned year : 7;
};
struct dos_time {
    unsigned second : 5;
    unsigned minute : 6;
    unsigned hour : 5;
};

/* Tables for date-time routines */
static mon_days[13]={
    0,31,59,90,120,151,181,212,243,273,304,334,365};
static char *day_names[7] = {
    "sunday", "monday", "tuesday", "wednesday", "thursday",
    "friday", "saturday"};
static char *mon_names[12] = {
    "january","february","march","april","may","june",
    "july","august","september","october","november","december"};
static char *ampm_table[] = {
    "am", "a.m.", "pm", "p.m."};
static char *ord_table[] = {
    "th", "st", "nd", "rd", "th", "th", "th", "th", "th", "th"};

/************************************************************************
 *			Do It						*
 ************************************************************************/
 
main()
{
    int set_it;
    union {
	long int Long;
	struct {
	    int date_w;
	    int time_w;
	} words;
	struct {
	    struct dos_date date_f;
	    struct dos_time time_f;
	} fields;
    } date_time;	/* Long word date/time for keyboard clock */

    /* First a little commercial */
    clr_scr();
    pr_strg("  \\AUTO\\ Folder Date & Time Setter\r\n");
    pr_strg("           Vanya J. Cooper\r\n");
    pr_strg("             7-Aug-1986\r\n\n\n");

    /* Check for a valid date and ask for new one if not */
    date_time.Long = Gettime();	/* Read the keyboard clock */
    if (date_time.fields.date_f.day == 0 || date_time.fields.date_f.month == 0
      || (date_time.fields.date_f.year + 1980) < THIS_YEAR) {
	rev_vid();
	pr_strg("Date:");
	nor_vid();
	pr_strg(" ");
	set_it = ask_date(&date_time.fields.date_f);
    }
    else
	set_it = FALSE;
    pr_strg("\rDate: ");
    era_eol();
    pr_daydate(date_time.fields.date_f.month, date_time.fields.date_f.day,
	date_time.fields.date_f.year + 1980);
    pr_strg("\r\n\n");
    if (set_it) {
	rev_vid();
	pr_strg("Time:");
	nor_vid();
	pr_strg(" ");
	ask_time(&date_time.fields.time_f);
    }
    pr_strg("\rTime: ");
    pr_time(date_time.fields.time_f.hour, date_time.fields.time_f.minute,
	date_time.fields.time_f.second * 2);
    era_eol();
    pr_strg("\r\n");
    if (set_it) {
	Settime(date_time.Long);		/* Set keyboard clock */
	Tsetdate(date_time.words.date_w);	/* Set DOS date */
	Tsettime(date_time.words.time_w);	/* Set DOS time */
	}
}

/*----------------------------------------------------------------------*
 * ask_date - ask user for the date					*
 * Returns: TRUE if date accepted					*
 *          FALSE if just <CR>						*
 *----------------------------------------------------------------------*/

ask_date(date_ptr)

struct dos_date *date_ptr;

{
    int day, month, year, len;
    char line_buf[MAXLINE];

    line_buf[0] = '\0';
    for (;;) {
	if (!(len = edit_line(line_buf, MAXLINE)))
	    return(FALSE);
	if (ss_date(line_buf, &month, &day, &year) == len)
	    if (year >= 1980) {
		date_ptr -> year = year - 1980;
		date_ptr -> month = month;
		date_ptr -> day = day;
		return(TRUE);
	    }
	/* Bad date */
	pr_char(CH_BEL);
    }
}

/*----------------------------------------------------------------------*
 *  ask_time - Ask user for the time					*
 *  Returns: TRUE if time accepted					*
 *           FALSE if just <cr> 					*
 *----------------------------------------------------------------------*/

ask_time(time_ptr)

struct dos_time *time_ptr;

{
    int hour, minute, second, len;
    char line_buf[MAXLINE];

    line_buf[0] = '\0';
    for (;;) {
	if (!(len = edit_line(line_buf, MAXLINE)))
	    return(FALSE);
	if (ss_time(line_buf, &hour, &minute, &second) == len) {
	    time_ptr -> hour = hour;
	    time_ptr -> minute = minute;
	    time_ptr -> second = second / 2;
	    return(TRUE);
	}
	/* Invalid time */
	pr_char(CH_BEL);
    }
}

/*----------------------------------------------------------------------*
 * edit_line - edit a string						*
 *									*
 * Call with string already displayed on screen and cursor positioned	*
 * just after end.  Returns the new number of characters in the buffer.	*
 *----------------------------------------------------------------------*/

edit_line(buf, bufsize)

char	buf[];		/* pointer to buffer */
int	bufsize;	/* max buffer size */

{
    int i;
    char c;

    i = strlen(buf);		/* initial length */
    while ((c = c_necin()) != '\r') {
	switch (c) {
	case '\b':
	case CH_DEL:
	    if (i > 0) {
		pr_strg("\b \b");
		--i;
	    }
	    else
		pr_char(CH_BEL);
	    break;
	case CH_ESC:
	    while (i > 0) {
		pr_strg("\b \b");
		--i;
	    }
	    break;
	default:
	    if (c >= ' ' && i < (bufsize - 1)) {
		pr_char(c);
		buf[i++]=c;
	    }
	    else
		pr_char(CH_BEL);
	}
    }
    buf[i] = '\0';	/* Mark end with NUL */
    return(i);
}

/************************************************************************
 * Routines to parse strings.       					*
 ************************************************************************/

/*----------------------------------------------------------------------*
 *  ss_date - Parse a date string					*
 *									*
 *  Expected Date formats:						*
 *    MM/DD[/[YY]YY]] | MM-DD[-[YY]YY] | MM DD[ [YY]YY] | MMDD[[YY]YY]	*
 *    DD-MONTH[-[YY]YY] | DD MONTH[ [YY]YY] | DDMONTH[[YY]YY]		*
 *    MONTH DD[st|nd|rd|th][, [YY]YY]					*
 *    MM is numeric month, MONTH is unique alpha month name		*
 *  Returns: number of characters processed				*
 *----------------------------------------------------------------------*/

ss_date(s, mon_ptr, day_ptr, yr_ptr)

char *s;				/* pointer to string to parse */
int *mon_ptr, *day_ptr, *yr_ptr;	/* return values */

{
    int len, i;
    char *p;

    p = s;
    *mon_ptr = *day_ptr = *yr_ptr = 0;
    p += ss_any(p, " ");		/* skip leading blanks */
    if (len = ss_dec(p, 2, mon_ptr)) {	/* assume MM DD YY */
	p += len;			/* advance pointer */
	set_lower(p, 2);		/* convert to lower case */
	if (strncmp(p, ord_table[*mon_ptr % 10], 2) == 0)
	    p += 2;			/* 1st, 2nd, etc */
	p += ss_any(p, "-/ ");		/* skip noise characters */
    }
    if (len = ss_alpha(p)) {		/* if month name coming */
	*day_ptr = *mon_ptr;		/* first number (if any) was day */
	set_lower(p, len);		/* convert to lower case */
	if (tbl_find(p, len, mon_names, 12, mon_ptr) == 1) {
	    (*mon_ptr)++;		/* months start at 1 */
	    p += len;			/* advance scan pointer */
	    p += ss_any(p, "-/ ");	/* skip more junk */
	}
    }
    if (!*day_ptr) {			/* if no day yet */
	p += ss_dec(p, 2, day_ptr);	/* try again */
	set_lower(p, 2);		/* convert to lower case */
	if (strncmp(p, ord_table[*day_ptr % 10], 2) == 0)
	    p += 2;			/* 1st, 2nd, etc */
	p += ss_any(p, ",-/ ");		/* Noise characters */
    }
    p += ss_dec(p, 0, yr_ptr);		/* Now get the year */
    p += ss_any(p, " ");		/* skip trailing blanks */
    if (!*yr_ptr)
        *yr_ptr = THIS_YEAR;		/* default is this year */
    else if (*yr_ptr < 100)		/* convert 2 digit year */
	*yr_ptr += 1900;		/* to 4 digits */
        
    /* Now check that this could be a real date */
    if(*mon_ptr && *day_ptr && *mon_ptr <= 12 ) {
	if (*day_ptr <= (mon_days[*mon_ptr] - mon_days[*mon_ptr - 1])
	  || ((*yr_ptr % 100) && !(*yr_ptr % 4)	/* leap year */
	  && *mon_ptr == 2 && *day_ptr == 29))
	    return(p - s);		/* success */
    }
    else
	return(0);			/* error return */
}

/*----------------------------------------------------------------------*
 *  ss_time - Parse a time string					*
 *  Time formats accepted:						*
 *    HH[:MM][:SS][AM|A.M.|PM|P.M.]					*
 *    HH[ MM][ SS][AM|A.M.|PM|P.M.]					*
 *    HH[MM][SS][AM|A.M.|PM|P.M.]					*
 *    if no AM/PM interpreted as 24 hour time				*
 *  Returns: Number or characters parsed				*
 *----------------------------------------------------------------------*/

ss_time(s, hr_ptr, min_ptr, sec_ptr)

char	*s;				/* string to parse */
int 	*hr_ptr, *min_ptr, *sec_ptr;	/* results */

{
    int len;
    char *p;

    p = s;				/* initial pointer */
    p += ss_any(p, " ");		/* leading spaces */
    p += ss_dec(p, 2, hr_ptr);		/* get hours */
    p += ss_any(p, ": ");		/* noise */
    p += ss_dec(p, 2, min_ptr);		/* get hours */
    p += ss_any(p, ": ");		/* noise */
    p += ss_dec(p, 2, sec_ptr);		/* get hours */
    p += ss_any(p, " ");		/* noise */
    if (len = ss_any(p, "AaPpMm.")) {	/* look for am or pm */
	int ampm;			/* temp variable */
	set_lower(p, len);		/* lowerize */
	if (tbl_find(p, len, ampm_table, 4, &ampm)) {
	    if (*hr_ptr < 1 || *hr_ptr > 12 )
		return(0);		/* out of range */
	    if (*hr_ptr == 12)		/* fix up hour */
		*hr_ptr = 0;
	    if (ampm >= 2)		/* pm only */
		*hr_ptr += 12;
	    p += len;			/* bump pointer */
	}
    }
    p += ss_any(p, " ");		/* skip trailing blanks */
    if (*hr_ptr < 24 && *min_ptr < 60 && *sec_ptr < 60)
	return(p - s);			/* success */
    else
	return(0);			/* failure */
}

/*----------------------------------------------------------------------*
 * ss_dec - scan string for decimal number				*
 *----------------------------------------------------------------------*/

ss_dec(s, len, num)

char	*s;	/* pointer to string */
int	len;	/* max length of number or 0 */
int	*num;	/* the number */

{
    int i;
    *num = i = 0;
    do {
	if (*s >= '0' && *s <= '9')
	    *num = *num * 10 + *s++ - '0';
	else
	    break;
    } while (++i != len);
    return(i);
}

/*----------------------------------------------------------------------*
 * ss_alpha - return length of next alpha part				*
 *----------------------------------------------------------------------*/

ss_alpha(s)

char *s;	/* the string */

{
    int i;
    for (i = 0;(*s >= 'A' && *s <= 'Z')
      || (*s >= 'a' && *s <= 'z');i++)
	s++;
    return(i);
}

/*----------------------------------------------------------------------*
 * ss_any - return size of string of arbitrary characters		*
 *----------------------------------------------------------------------*/

ss_any(s, list)

char s[];	/* pointer to input string */
char list[];	/* pointer to list of chars to look for */

{
    int i,j;
    i = j = 0;		/* initialize counts */
    while (list[j])
	if (s[i] == list[j++]) {
	    j = 0;	/* start list over */
	    i++;	/* count characters */
	}
    return(i);
}


/************************************************************************
 *		Output routines						*
 ************************************************************************/

/*----------------------------------------------------------------------*
 * pr_daydate - print day of the week and date				*
 *----------------------------------------------------------------------*/

pr_daydate(month, day, year)

int month, day, year;

{
    int wday;
    wday = weekday(month, day, year);
    pr_capword(day_names[wday]);
    pr_strg(", ");
    if (month > 0 && month <= 12)
	pr_capword(mon_names[month - 1]);
    else
	pr_0dec(0, month);			/* bad name */
    pr_char(' ');
    pr_0dec(0,day);
    pr_strg(ord_table[day % 10]);
    pr_strg(", ");
    pr_0dec(0, year);
}

/*----------------------------------------------------------------------*
 * pr_time - print a time						*
 *----------------------------------------------------------------------*/

pr_time(hour, minute, second)

int hour, minute, second;

{
    if (hour == 0)
	pr_0dec(0, 12);
    else if (hour < 13)
	pr_0dec(0, hour);
    else
	pr_0dec(0, hour - 12);
    pr_char(':');
    pr_0dec(2, minute);
    pr_char(':');
    pr_0dec(2, second);
    if (hour < 12)
	pr_strg(" am");
    else
	pr_strg(" pm");
}

/*----------------------------------------------------------------------*
 * pr_0dec - print leading zero decimal number				*
 *----------------------------------------------------------------------*/

pr_0dec(size, num)

int size;	/* field size or 0 */
int num;	/* the number */

{
    if (size > 1 || num > 9) 
	pr_0dec(size - 1, num / 10);
    pr_char((num % 10) + '0');
}

/*----------------------------------------------------------------------*
 * pr_capword - print a capitalized word				*
 *----------------------------------------------------------------------*/

pr_capword(s)

char	*s;

{
    if (*s >= 'a' && *s <= 'z')
        pr_char(*s++ - ('a' - 'A'));
    pr_strg(s);
}

/************************************************************************
 * Miscellaneous Routines						*
 ************************************************************************/

/*----------------------------------------------------------------------*
 * tbl_find - lookup string in table					*
 * Returns: number of possible matches					*
 *----------------------------------------------------------------------*/

tbl_find(p, len, names, tlen, item)

char	*p;		/* pointer to string */
int	len;		/* length of source string */
char	*names[];	/* char pointer array of names */
int	tlen;		/* length of table */
int	*item;		/* item number return */

{
    int n, m;

    for(n = m = 0;m < tlen;m++)		/* search table */
    if (strncmp(p, names[m], len) == 0) {
	*item = m;			/* store item number */
	n++;				/* count matches */
	if (strlen(names[m]) == len)	/* if exact match */
	    return(1);			/* give good return */
    }
    return(n);				/* return number matches */
}

/*----------------------------------------------------------------------*
 *  weekday - get day of week for a date				*
 *  Returns: 0 = Sunday ... 6 = Saturday				*
 *----------------------------------------------------------------------*/

weekday(month, day, year)

int	month, day, year;

{
    int w;

    if(year < 1 || month < 1 || month > 12 || day < 1)
	return(0);			/* error */
    w = (year - 1) * 365 + (year - 1) / 4 - (year - 1) / 100
	+ mon_days[month - 1] + day + 5;
    if ((year % 100) && !(year % 4) && month > 2)
	w++;			/* correct for leap year */
    return(w % 7);
}

/*----------------------------------------------------------------------*
 * set_lower - set string to lower-case					*
 *----------------------------------------------------------------------*/

set_lower(s, len)

char	*s;	/* string pointer */
int	len;	/* maximum len or 0 */

{
    do {
	if (*s >= 'A' && *s <= 'Z')
	    *s += 'a' - 'A';
    } while (--len && *s++);
}

/*----------------------------------------------------------------------*
 * strlen - return length of a string					*
 *----------------------------------------------------------------------*/

strlen(s)

char *s;

{
    int i;
    for (i=0;*s++;i++)
	;
    return(i);
}

/*----------------------------------------------------------------------*
 * strncmp - compare strings up to a length				*
 *----------------------------------------------------------------------*/

strncmp(s1, s2, len)

char *s1, *s2;
int len;

{
    while (--len > 0 && *s1 && *s1 == *s2)
	s1++, s2++;
    return(*s1 - *s2);
}

/* begin comment

***********************************************************************
*   QSTART.S - Assembler startup module                               *
*              Extract and assemble separately                        *
***********************************************************************

*	QSTART - Quick start for C programs
*	Only sets up stack, cannot use GEMLIB, AESBIND or VDIBIND
*	Must be first object file in link statement
*
*	Vanya J. Cooper
*	5-Jul-1986
*

	.text
	.globl	_main

	move.l	#ustk,a7	* get program stack
	jsr	_main		* go to program
	move.l	#0,(a7)		* back to gemdos
	trap	#1

*	Allocate space for stack

	.bss
	.even
	.ds.l	500
ustk:	.ds.l	1

	.end

end comment */
