/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> BOOT.C  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
> This program reboots the operating system.
> Setting the system to 1 of 5 different modes.
> Each mode is custom setup by the User.
> The User defines which accessory to be included on the desktop.
> The User defines where the operating system will be booted from.
> The User defines which memory size (512K or 1 Meg) to setup.
>
> Written by Steve Clark.
> of
> Nebula Software.
>>>>>>>>>>>>>>>>>>>>>>>>>>>>>  BOOT.C   <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/


/*----- INCLUDES  -----*/

#include <gemdefs.h>
#include <obdefs.h>
#include <stdio.h>
#include <osbind.h>
#include <boot.h>


/*----- DEFINES ------*/

#define	  FILE_SIZE    400		/*  data file size */
#define   MAX_FILES    8		/*  maximum number of accessories */


/*----- GLOBAL VARIABLES -----*/

int     intin[128],intout[128];		/*  GEM bindings */
int     contrl[12],ptsin[128],ptsout[128];
int	message_buf[8];			/*  evnt_multi message buffer */
int	old_mode;			/*  copy of old selected mode */
char	data_file[15] = "A:BOOT.DAT";	/*  data filename  */
char	desktop[15] = "A:DESKTOP";	/*  part of desktop filename */
char	tos[16] = "A:TOS";		/*  part of operating system filename */
long	menu_add;			/*  address of MENU bar object  */

struct  {				/*  this structure is saved as the 
					/*  data file  */

	int	op_system;		/*  ROM = 1, RAM = 0 */
	int	memory_size;		/*  1MEG = 1, 512K = 0 */
	int	file_cnt;		/*  number of accessories  */
	int	selected_mode;		/*  the last mode before reboot  */
	int	acc_status[10];		/*  array of objects in message1  */
	int	acc_line[10];		/*  array of objects in message1  */
	int	mode_handle;		/*  object ptr to be check marked  */
	int	*mode_ptr[5];		/*  array of ptrs to mode data arrays */
	int	mode_data[60];		/*  array for mode data  */
	char	*acc_fil_ptr[10];	/*  array of ptrs to each accessory  */
	char	access_files[170];	/*  array containing acc. filenames */

	}  data;



/********************************/
/*    MAIN PROGRAM              */
/********************************/


main()

{
	appl_init();		/*  REGISTER WITH GEM  */
	principal();		/*  BODY OF PROGRAM    */
	end();                  /*  TERMINATE PROGRAM  */
}



/*---------------------|----------------------------------------------------<
>     principal loop   |  Main loop of the program            		    <
>----------------------|----------------------------------------------------*/

principal()

{
int	abort,event,dummy,result,file_handle;
long	savessp,dum,count;


	rsrc_load("BOOT.RSC");			/*  load resource file  */
	rsrc_gaddr(0,MENU,&menu_add);		/*  get address of MENU */
	result = Fsfirst(data_file,0);		/*  see if data file on disk */
	if ( result < 0)set_menu(0);		/*  if No, diable menu items */
	else					/*  else data file is there  */
	   {
		menu_ienable(menu_add,INIT,0);	/*  disable INIT   */
		file_handle = Fopen(data_file,2);  /*  open file   */
		if(file_handle > 0)		   /* if open success */
		  {
			count = FILE_SIZE;	   /*  read file & close  */
			count = Fread(file_handle,count,&data.op_system);
			Fclose(file_handle);
		  }
		set_pointrs();			/*  setup pointers  */
		menu_icheck(menu_add,MODE_1,0);
		menu_icheck(menu_add,data.mode_handle,1);
		old_mode = data.selected_mode;		/* copy old mode */
	    }

	menu_bar(menu_add,1);			/*  turn on menu  */
	abort = 0;				/*  initialize  abort flag */

	do 			
	  {
		event = evnt_multi(MU_MESAG | MU_TIMER,
 		  0,0,0,
   		  0,0,0,0,0,			/*  this loop looks for a - */
		  0,0,0,0,0,			/*  message telling of a - */
		  message_buf,			/*  menu selection .  It is */
		  1000,0,			/*  interrupted every 1/10th */
		  &dummy,&dummy,		/*  of a second to set the   */
		  &dummy,			/*  mouse pointer to the hand */
		  &dummy,			/* , this is in case an acc-  */
		  &dummy,			/*  essory was selected & it */
		  &dummy);			/*  changed the mouse form.  */

		wind_update(1);			/* I'll take over the screen. */
		graf_mouse(3,0L);		/*  set mouse form to Hand  */
			/*  this loop is done if a menu item was selected   */
		if((event & MU_MESAG) && (message_buf[0] & MN_SELECTED))
		  {
			menu_tnormal(menu_add,message_buf[3],1);
			switch (message_buf[4] )
			 {
				case INFO:		/* was it Desk Info ? */
					do_info();
					break;
				case INIT:		/* was it Initialize */
					do_init();
					break;
				case DEFIN_MD:		/* was it Define ?  */
					define();
					break;
				case MODEINFO:		/* was it ModeInfo ? */
					mode_info();
					break;
				case REBOOT:		/* was it Reboot ? */
					result = reboot();
					abort = result;
					break;
				case ABORT:		/* was it Abort ? */
					abort = 1;
					break;
				case MODE_1:		/* was it Mode_1 ? */
				     menu_icheck(menu_add,data.mode_handle,0);
				     data.selected_mode = 0;
				     data.mode_handle = MODE_1;
				     menu_icheck(menu_add,data.mode_handle,1);
					break;
				case MODE_2:		/* was it Mode_2 ? */
				     menu_icheck(menu_add,data.mode_handle,0);
				     data.selected_mode = 1;
				     data.mode_handle = MODE_2;
				     menu_icheck(menu_add,data.mode_handle,1);
					break;
				case MODE_3:		/* was it Mode_3 ? */
				     menu_icheck(menu_add,data.mode_handle,0);
				     data.selected_mode = 2;
				     data.mode_handle = MODE_3;
				     menu_icheck(menu_add,data.mode_handle,1);
					break;
				case MODE_4:		/* was it Mode_4 ? */
				     menu_icheck(menu_add,data.mode_handle,0);
				     data.selected_mode = 3;
				     data.mode_handle = MODE_4;
				     menu_icheck(menu_add,data.mode_handle,1);
					break;
				case MODE_5:		/* was it Mode_5 ? */
				     menu_icheck(menu_add,data.mode_handle,0);
				     data.selected_mode = 4;
				     data.mode_handle = MODE_5;
				     menu_icheck(menu_add,data.mode_handle,1);
					break;
			    }   /*   end of switch  */

		  }  /*  end  of  if  */

		wind_update(0);		/*  GEM can have the screen  */
	    } while ( !abort );
	graf_mouse(0,0L);		/* set the mouse form to arrow */

}


/*-----------------------------|---------------------------------------------<
>     do_info                  | Display the Desk Info box                   <
>------------------------------|--------------------------------------------*/

do_info()

{
OBJECT	*dialog_tree;
int	cx,cy,cw,ch,button;

	rsrc_gaddr(0,INFO_BX,&dialog_tree);	/*  Get address of Info box */

	form_center(dialog_tree,&cx,&cy,&cw,&ch);  /* find the coordinates  */
	form_dial(FMD_START,cx,cy,cw,ch,0,0,0,0);  /* reserve space for it  */
	objc_draw(dialog_tree,0,3,cx,cy,cw,ch);	   /*  Draw it              */
	button = form_do(dialog_tree,0);	   /*  Let GEM do the work  */
	form_dial(FMD_FINISH,cx,cy,cw,ch,cx,cy,cw,ch);/*  Free space & erase*/
	dialog_tree[EXIT_INF].ob_state &= ~SELECTED;	/* deselect Button  */

}


/*--------------------|--------------------------------------------------------<
>    set_pointrs      |  Initialize pointers arrays to point to proper areas   <
>                     |   inside the structure  data .                         <
>---------------------|-------------------------------------------------------*/

set_pointrs()

{
int	buf_cnt,a;

    /*  Initialize array of pointers to accessory names  */

	buf_cnt = 0;
	for ( a = 0; a < MAX_FILES; a++)
	   {
		data.acc_fil_ptr[a] = &data.access_files[buf_cnt];
		buf_cnt += 16;
	    }			/*  16 bytes for each filename * number
					of files  */

    /*  Initialize array of pointers to mode data  segments  */

	buf_cnt = 0;		/*  start buffer count at 0  */
	for (a = 0;  a < 5;  a++)
	  {
		data.mode_ptr[a] = &data.mode_data[buf_cnt];
		buf_cnt +=12;
	    }			/*  60 int's in all , with 5 modes with 
					12 variables in each  */
}


/*-----------------------|-----------------------------------------------------<
>     do_init            |   Must not be a data file on disk , so initialize   <
>                        |    everything to simulate starting with mode 1.     <
>------------------------|----------------------------------------------------*/

do_init()

{
char	line1[26],line2[26],line3[26],line4[26];
int	ptr_cnt,result,a,but,dummy;

  /*  Find out if User's system is in ROM ?  */

	strcpy(&line1," ");
	strcpy(&line2,"Does your Atari ST ");
	strcpy(&line3,"have TOS on Rom ?");
	strcpy(&line4," ");
	data.op_system = show_mesg0(&line1,&line2,&line3,&line4);

   /*   Find out if User's system is 1 meg ?  */

	strcpy(&line2,"Does your ST have 1-Meg ");
	strcpy(&line3," ");
	data.memory_size = show_mesg0(&line1,&line2,&line3,&line4);

	ptr_cnt = 0;			/*  start pointer count at 0 */
	data.file_cnt = 0;		/*  start file count at 0  */
	data.selected_mode = 0;		/*  set Initial mode to mode 1  */
	data.mode_handle = MODE_1;	/*  save info on checked item  */
	set_pointrs();			/*  go setup pointers  */
	save_status();			/*  save object  identifiers  */

   /*  loop to find the accessories to be used for User's selection */

	do 
	  {
		result = find_fname(data.acc_fil_ptr[ptr_cnt]);	
		data.file_cnt++;
		if(!result) data.file_cnt--;
		if(data.file_cnt == MAX_FILES) result = 0;
		ptr_cnt++;
		do        /*  debounce mouse button  */
		 {
			a = graf_mkstate(&dummy,&dummy,&but,&dummy);
		 } while (but);
	   }  while (result );

     /*   Clear all mode data  */

	for (a = 0;  a < 60; a++) data.mode_data[a] = -1 ;


	menu_bar(menu_add,0);		/*  erase menu bar  */
	menu_ienable(menu_add,INIT,0);	/*  disable  Initialize in menu bar  */
	set_menu( 1 );			/*  enable everything else in menu   */
	menu_bar(menu_add,1);		/*  show menu bar  */

}


/*-----------------------|----------------------------------------------------<
>     save_status        |  Save a copy of object pointer (used in Mode Info) <
>                        |    in structure data                               <
>------------------------|---------------------------------------------------*/

save_status()

{

  /*   Copy object pointers for value (used in Mode Info)  */

	data.acc_status[0]  =  CK0;
	data.acc_status[1]  =  CK1;
	data.acc_status[2]  =  CK2;
	data.acc_status[3]  =  CK3;
	data.acc_status[4]  =  CK4;
	data.acc_status[5]  =  CK5;
	data.acc_status[6]  =  CK6;
	data.acc_status[7]  =  CK7;
	data.acc_status[8]  =  CK8;
	data.acc_status[9]  =  CK9;

   /*   Copy object pointers for subject line  (used in Mode Info)  */

	data.acc_line[0]  =  MSTR_0;
	data.acc_line[1]  =  MSTR_1;
	data.acc_line[2]  =  MSTR_2;
	data.acc_line[3]  =  MSTR_3;
	data.acc_line[4]  =  MSTR_4;
	data.acc_line[5]  =  MSTR_5;
	data.acc_line[6]  =  MSTR_6;
	data.acc_line[7]  =  MSTR_7;
	data.acc_line[8]  =  MSTR_8;
	data.acc_line[9]  =  MSTR_9;

}


/*-----------------------|-----------------------------------------------------<
>                        |  Routine to display dialog box message0.            <
>     show_mesg0         |  Calling routines pass 4 pointers to strings that   <
>                        |   are to be displayed inside of box. Returns which  <
>                        |   button was selected upon exit.                    <
>------------------------|----------------------------------------------------*/

show_mesg0(line1,line2,line3,line4)

char	*line1,*line2,*line3,*line4;

{
OBJECT	*tree;
int	cx,cy,cw,ch,button;


	rsrc_gaddr(0,MESSAGE0,&tree);		/*  get address of box  */

	tree[STRING1].ob_spec = line1;   	/*  pass pointer of line 1 */
	tree[STRING2].ob_spec = line2;   	/*  pass pointer of line 2 */
	tree[STRING3].ob_spec = line3;   	/*  pass pointer of line 3 */
	tree[STRING4].ob_spec = line4;   	/*  pass pointer of line 4 */
	form_center(tree,&cx,&cy,&cw,&ch);	/*  Find center coordinates */
	form_dial(FMD_START,cx,cy,cw,ch,0,0,0,0);  /* Reserve space  */
	objc_draw(tree,0,2,cx,cy,cw,ch);	/*  draw  box  */
	button = form_do(tree,0);		/*  let GEM do the work  */
	form_dial(FMD_FINISH,cx,cy,cw,ch,cx,cy,cw,ch);  /*  Free space & erase*/

    /*   determine which button was pressed & deselect  */

	if(button == YES_BUT)
	  {
		tree[YES_BUT].ob_state  &= ~SELECTED;
		button = 1;
	   }
	else
	   {
		tree[NO_BUT].ob_state  &= ~SELECTED;
		button = 0;
	    }

	return ( button );		/*  return selected button  info */

}

/*---------------------|-------------------------------------------------------<
>     find_fname       |   Routine which calls directory dialog window.        <
>                      |   It is passed a pointer to a string to copy a file-  <
>                      |    name to.                                           <
>----------------------|------------------------------------------------------*/
	
find_fname(f_sel)

char	*f_sel;

{
char	temp[20],f_path[14];
int	error,fbut,len;

	do
	  {
		f_sel[0] = '\0';		/*  Clear File selection */
		strcpy(&f_path, "A:*.*");	/*  Copy File path   */
		error = fsel_input(&f_path,f_sel,&fbut);   /* Call window */

		if(error > 0 && fbut >=1)	/* If no error and OK selected*/
		    {
		/*  Then build filename string */
			strncpy(&temp,&f_path,2);
			temp[2] = '\0';
			strcat(&temp,f_sel);
			strcpy(f_sel,&temp);
		     }

       /*  try to catch filename of blanks & redo  */
		if((len = strlen(f_sel)) < 4 && fbut) error = -1;

	   }  while(error < 0);

	return(fbut);		/* Return which button was selected  */

}  /* end  find_fname()  */

/*-----------------|-----------------------------------------------------------<
>   set_menu       |   Routine to handle enabling or disabling most menu items <
>------------------|----------------------------------------------------------*/

set_menu(value)

int	value;

{

	menu_ienable(menu_add,DEFIN_MD,value);
	menu_ienable(menu_add,MODEINFO,value);
	menu_ienable(menu_add,REBOOT,value);
	menu_ienable(menu_add,MODES,value);
}


/*--------------------|-------------------------------------------------------<
>      reboot         |  Routine to change neccessary value and cause a reset <
>---------------------|------------------------------------------------------*/

reboot()

{
char	line1[25],line2[25],line3[25],line4[25];
char	tmp[2];
int	result,file_handle,error1,error2,error3;
int	fatal,error0,error4;
long	count,dummy;

	fatal = 0;	/*   reset fatal flag  */
   /*  display last chance  warning  */

	strcpy(&line1,"   ****  WARNING  *****");
	strcpy(&line2,"REBOOTING MODE ");
	tmp[0] = 0x30 + (data.selected_mode + 1);
	tmp[1] = '\0';
	strcat(&line2,&tmp);
	strcat(&line2," will");
	strcpy(&line4,"[Yes] continue [No] abort");
	strcpy(&line3,"ERASE Data in RAM !!");
	error0 = show_mesg0(&line1,&line2,&line3,&line4);

   /*   if last chance was continue , then charge  */

	if ( error0)
	  {
		/*  save  all of Boot data in a data file  */

		file_handle = Fcreate(data_file,0);
		if (file_handle > 0)
		   {
			count = FILE_SIZE;
			count = Fwrite(file_handle,count,&data.op_system);
			Fclose(file_handle);
			error4 = 1;	/*  flag for save OK  */
		    }
		else	/*  display error message if could not open file */
		   {
			strcpy(&line1,"Could not Save Data File");
			strcpy(&line2,"Press any key to resume");
			strcpy(&line3,"Data file not saved !! ");
			line4[0] = '\0';
			show_mesg0(&line1,&line2,&line3,&line4);
			fatal = 0;	/*  not a fatal error */
			error4 = 0;	/*  flag for change nothing */
		    }	
		if(error4)	/*  determine if save was a success */
		   {		/*  if yes continue */
			error1 = rename();
			error2 = set_desktop_inf();
			if (data.memory_size )  set_memsize();
			if (data.op_system)  error3 = set_op_sys();
			else  error3 = 0;
		    }

		/*  if all went well let's reset the system  */

		if( error0 && error4 &&!error1 && !error2 && !error3 )
		  {
			count = Super(0L);
	
			asm  {
			movea.l		#0x420, A0	/* set A0 -> memvalid */
			move.l		#0x752019f3, (A0)  /* put magic # in  */
			move.l		#4, A0	   /*  set A0 -> reset add */
			move.l		(A0), A1   
			jmp		(A1)	   /* jump to reset address */
			      }     /* if everything works, should never 
					get past here */

			dummy = Super(count);  /*  just in case return to  */
		   }				/*   to User mode  */

		else  fatal = 1;   /* if error in setup of reset set flag */

	  }  /*  end of if last chance  */

	return(fatal);	/*  if unsuccessful return status  */

}


/*---------------------|------------------------------------------------------<
>    set_op_sys        |  Routine to rename TOS if neccessary.                <
>----------------------|-----------------------------------------------------*/

set_op_sys()

{
char	line1[25],line2[25],line3[25],line4[25];
char	temp[16],ext[5],dta_buf[44],old_file[16];
int	*mode_data,result,fatal;
long	old_buf;

	fatal = 0;		/*  clear fatal flag  */
	old_buf = Fgetdta();	/*  save old DTA address */
	Fsetdta(&dta_buf);	/*  create new DTA buffer */
     /*  Initialize tempory mode data pointer  */

	mode_data = data.mode_ptr[data.selected_mode];

	strcpy(&temp,&tos); 	/*  build path name  */
	strcat(&temp,".*");
	result = Fsfirst(&temp,0);  /*  see if it is on disk  */

	if ( result < 0)	/* if not, display error message */
	  {
		strcpy(&line1,"Cannot locate TOS ");
		strcpy(&line2,"on Drive A ");
		line3[0] = '\0';
		line4[0] = '\0';
		show_mesg0(&line1,&line2,&line3,&line4);
		fatal = 1;	/* set fatal flag */
	  }
	else 	/*  Or if successful */
	  {
		strcpy(&old_file,"A:");		/*  build old filename  */
		strcat(&old_file,&dta_buf[30]);
		strcpy(&temp,&tos);		/*  build new filename  */
		if (mode_data[10]) strcat(&temp,".OFF");
		else strcat(&temp,".IMG");
		result = Fsfirst(&temp,0);	/*  see if new filename is
						   on the disk */
		/*  if not then rename TOS */
		if ( result < 0) Frename(0,&old_file,&temp);
	   }

	Fsetdta(old_buf);	/*  restore old DTA address  */
	return(fatal);		/* return operation status  */
}

			
		  
/*---------------------|-------------------------------------------------------<
>   set_memsize        |  Routine to set the memory size variables to desired  <
>                      |   size according to mode data on selected mode        <
>----------------------|------------------------------------------------------*/

set_memsize()

{
long	savessp,source,dummy;
long	phystop,memtop;
int	*mode_data,banks;

	/*  initialize tempory pointer to mode data of selected mode */

	mode_data = data.mode_ptr[data.selected_mode];     

	if (mode_data[11] == 0)		/*  if 512K selected */
	  {
		banks = 0x0400;		/*  set banks data for 512K ,bank 1 */
		phystop = 0x80000;	/*  set top of memory for 512K */
		memtop = 0x78000;	/*  set top of usable mem. 512K */
	   }
	else	/*  Must be set for 1 Meg */
	   {
		banks = 0x0500;		/*  set banks data for 512K 2 banks */
		phystop = 0x100000;	/*  set top of memory to 1 meg */
		memtop = 0xf8000;	/*  set top of usable mem. 1 meg */
	    }

	savessp = Super(0L);		/*  goto Supervisory mode */

	asm {
		movea.l		#0x424, A0	  /* make A0 -> memcntlr */
		move.w		banks(A6), (A0)	  /*  put bank data there */
		movea.l		#0x42e, A0	  /*  make A0 -> Phystop */
		move.l		phystop(A6), (A0) /*  put phystop data there */
		movea.l		#0x436, A0	  /*  make A0 -> memtop  */
		move.l		memtop(A6), (A0)  /*  put memtop data there */
	    }

	dummy = Super(savessp);		/* return to User mode */
	Vsync();		/* This is not really nessary */
	source = Physbase();	/* Find screen's physical location */
	/* if the screen's physical location is not correct, change it */
	if (source != memtop) Setscreen(memtop,memtop,-1);
       
}


/*---------------------|-------------------------------------------------------<
>     mode_info        |  Routine displays window with Selected mode's data.   <
>----------------------|------------------------------------------------------*/

mode_info()

{
OBJECT	*dialog;
int	cx,cy,cw,ch,result,button;
int	*acc_value,a,b;
char	*value,mode_num[4],file[10];

	rsrc_gaddr(0,MESSAGE1,&dialog);		/*  get box address */
	/*  Initialize temporary pointer to selected mode data */
	acc_value = data.mode_ptr[data.selected_mode];
	/*  build mode number string */
	mode_num[0] = 0x30 + (data.selected_mode + 1);
	mode_num[1] = '\0';
	dialog[MES1INFO].ob_spec = mode_num;	/*  -> object to it */

	dialog[data.acc_line[0]].ob_spec = "TOS";  /*  setup line 1 */
	/*  determine if this mode has selected ROM or RAM */
	if (data.op_system && acc_value[10])
	     dialog[data.acc_status[0]].ob_spec = "ROM";
	else  dialog[data.acc_status[0]].ob_spec = "RAM";
	dialog[data.acc_line[1]].ob_spec = "RAM Size ";  /*  setup line 2 */
	/*  determine selected memory size for this mode */
	if (data.memory_size && acc_value[11] )
	      dialog[data.acc_status[1]].ob_spec = "1Mg";
	else dialog[data.acc_status[1]].ob_spec = "512";
	b = 0;		/*  Clear b */
	/*  This loop sets up remainding lines with values */
	for (a = 2; a < 10; a++)
	   {
		dialog[data.acc_line[a]].ob_spec = data.acc_fil_ptr[b];
		if(acc_value[b] == 0) value = "Off";
		else  if(acc_value[b] == 1) value = "On";
		else  value = " ";
		dialog[data.acc_status[a]].ob_spec = value;
		b++;
	     }
	/*  Standard display & deseleting here */
	form_center(dialog,&cx,&cy,&cw,&ch);
	form_dial(FMD_START,cx,cy,cw,ch,0,0,0,0);
	objc_draw(dialog,0,4,cx,cy,cw,ch);
	button = form_do(dialog,0);
	form_dial(FMD_FINISH,cx,cy,cw,ch,cx,cy,cw,ch);
	dialog[EXITMES1].ob_state &= ~SELECTED;

}

/*--------------------|-----------------------------------------------------<
>    define           |  Routine to take in User info for the selected mode <
>---------------------|----------------------------------------------------*/

define()

{
OBJECT	*dialog;
int	cx,cy,cw,ch,button;
int	a,b,c,*mode_data;
char	mode_num[2];
char	line1[25],line2[25],line3[25],line4[25];


	rsrc_gaddr(0,DEFINE_B,&dialog);    /*  get box address */
	form_center(dialog,&cx,&cy,&cw,&ch);	/* Find center coordinates */
	/*  Initialize temporary mode data pointer to selected mode */
	mode_data = data.mode_ptr[(data.selected_mode)];
	/*  build mode number string */
	mode_num[0] = 0x30 + (data.selected_mode + 1);
	mode_num[1] = '\0';
	dialog[MODENUM].ob_spec = mode_num;
	/*  Loop to display dialog box with accessory & take in User Info */
	for( a = 0;  a < data.file_cnt;   a++)
	  {
		dialog[FILENAME].ob_spec = data.acc_fil_ptrs[a];
		form_dial(FMD_START,cx,cy,cw,ch,0,0,0,0);
		objc_draw(dialog,0,4,cx,cy,cw,ch);
		button = form_do(dialog,0);
		form_dial(FMD_FINISH,cx,cy,cw,ch,cx,cy,cw,ch);
		if(button == A_BUT)
		  {
			 dialog[A_BUT].ob_state &= ~SELECTED;
			button = 1;
		   }
 		else 
		   {
			 dialog[B_BUT].ob_state  &= ~SELECTED;
			button = 0;
		   }
		mode_data[a] = button;
	   }
	/*  If Rom based system check User's prefference */
	if(data.op_system)
	  {
		strcpy(&line1,"Where do you want TOS ?");
		strcpy(&line2,"Press [ Yes ] for ROM.");
		strcpy(&line3,"Press [ No ] for RAM.");
		line4[0] = '\0';
		mode_data[10] = show_mesg0(&line1,&line2,&line3,&line4);
	   }
	/*  If Ram size is 1 meg check User's prefference */
	if ( data.memory_size )
	  {
		strcpy(&line1,"What size memory ?");
		strcpy(&line2,"Press [ Yes ] for 1-Meg.");
		strcpy(&line3,"Press [ No ] for 512K.");
		line4[0] = '\0';
		mode_data[11] = show_mesg0(&line1,&line2,&line3,&line4);
	  }

}



/*------------------|-------------------------------------------------------<
>  rename           |  Routine to rename accessories according to mode data.<
>-------------------|------------------------------------------------------*/

rename() 

{
long	dta_add;
int	a,b,period,error1,*mode_data,result;
int	fatal,i,length;
char	buffer[44],temp[20],file[20],*ext,*data_file;
char	line1[25],line2[25],line3[25],line4[25];


	dta_add = Fgetdta();		/*  get old DTA address */
	Fsetdta(&buffer);		/* Setup new DTA address */
	/*  Initialize temporary mode data pointer to selected mode data */
	mode_data = data.mode_ptr[data.selected_mode];
	fatal = 0;			/* Clear fatal flag */
	/*  Loop to check & rename if neccessary all accessories */
	for(a = 0; a < data.file_cnt; a++)
	  {
		strcpy(&temp,data.acc_fil_ptr[a]);  /*  Copy accessory name */
		/*  Find period in accessory name */
		length = strlen(&temp);	
		for(b =0; b <= length; b++)
			if(temp[b] == '.') period = b;
		if(period < length) temp[period] = '\0';
		else period = length + 1;
		strcat(&temp,".*");		/*  Build path name */
		error1 = Fsfirst(&temp,0);	/*  see if file is on disk */
		if( !error1)
		 {	/*  If yes , */
			strcpy(&file,"A:");	/*  build correct filename */
			strcat(&file,&buffer[30]);
			if (mode_data[a] == 1) ext = ".ACC";
			else  ext = ".OFF";
			temp[period] = '\0';
			strcat(&temp,ext);
			result = strcmp(&file,&temp);  /*  is file on disk 
							  correct ?  */
  /* if not rename */	if ( result != 0 ) result = Frename(0,&file,&temp);
		  }
		if(error1 < 0 )		/* if file was not on disk at all */
		  {			/*  display error message */
			strcpy(&line2," File I/O  Error !! ");
			strcpy(&line3," Reboot  Aborting ! ");
			line1[0] = '\0';
			line4[0] = '\0';
			show_mesg0(&line1,&line2,&line3,&line4);
			a = 11;		/* setup terminating count */
			fatal = 1;	/*  set fatal flag */
		   }
		
	    }  /*  end for loop  */
	Fsetdta(dta_add);	/*  Restore DTA address */
	return(fatal);		/*  Return operation status */
}

/*---------------------|------------------------------------------------------<
>   set_desktop_inf    |  Routine to rename Desktop Info file.                <
>----------------------|-----------------------------------------------------*/

set_desktop_inf()

{
char	ext[5],dta_buf[44],inf[5],temp[20];
char	line1[25],line2[25],line3[25],line4[25];
int	result,fatal;
long	savessp;

	savessp = Fgetdta();	/*  Save old DTA address */
	Fsetdta(&dta_buf);	/*  Setup new one */
	fatal = 0;		/*  Clear fatal flag */
	/*  if the old mode is not the same as the selected mode */

	if (old_mode != data.selected_mode)
	  {
		strcpy(&ext,".MD");	/*  Build filename for the */
		ext[3] = 0x30 + (old_mode + 1);  /*  selected mode */
		ext[4] = '\0';
		strcpy(&temp,&desktop);
		strcat(&temp,&ext);
		result = Fsfirst(&temp,0);  /* See if it is on the disk */
		if (result == 0)	/* If yes, something is wrong */
		  {		/*  so display an error message */
			strcpy(&line1,"CONFLICT : ");
			strcpy(&line2,"The old desktop file ");
			strcpy(&line3,"already exists !");
			show_mesg0(&line1,&line2,&line3,&temp);
			fatal = 1;
		  }
		if (result < 0)		/* if not, continue */
		  {
			strcat(&desktop,".INF");  /*  Build standard filename */
			result = Fsfirst(&desktop,0);  /* is it on the disk */
			if (result < 0)	/*  if not something is wrong */
			 {       /* So display an error message */
				strcpy(&line1,"Cannot find : ");
				line2[0] = '\0';
				line4[0] = '\0';
				show_mesg0(&line1,&line2,&desktop,&line4);
				fatal = 1;
			  }
			if ( result == 0)   /* If yes then continue */
			  {	/* Rename DESKTOP.INF to old mode */
				result = Frename(0,&desktop,&temp);
/*  build new mode filename */	temp[12] = 0x30 + (data.selected_mode + 1);
/*  is it on the disk */	result = Fsfirst(&temp,0);
/*  If yes rename it */		if (result == 0)  Frename(0,&temp,&desktop);
				else	/* if not then error */ 
				  {
					strcpy(&line2,
					   "Could not locate file : ");
					line1[0] = '\0';
					line4[0] = '\0';
					show_mesg0(&line1,&line2,&temp,&line4);
					fatal = 1;
				   }

			 }

		   }
	    }
	/*  If the mode hasn't changed I'll check things out anyway */
	if (old_mode == data.selected_mode)
	   {		/* Build strings */
		strcpy(&temp,&desktop);
		strcat(&temp,".MD");
		ext[0] = 0x30 + (old_mode + 1);
		ext[1] = '\0';
		strcat(&temp,&ext);
		result = Fsfirst(&temp,0); /* is the old mode file deselected */
		if (!result)	/*  if yes, display error */
		  {
			strcpy(&line1,"ERROR - desktop file");
			strcpy(&line2,"for this mode is disabled");
			strcpy(&line3,"Press any key to abort");
			line4[0] = '\0';
			show_mesg0(&line1,&line2,&line3,&line4);
			fatal = 1;
		   }
		else	/*  if not make sure standard file is there */
		 {
			temp[10] = '\0';
			strcat(&temp,"INF");
			result = Fsfirst(&temp,0); 
			if (result < 0)		/* if not display error */
			  {
				strcpy(&line1,"Cannot find the file:");
				line3[0] = '\0';
				strcpy(&line4,"Press any key to abort !");
				show_mesg0(&line1,&temp,&line3,&line4);
				fatal = 1;
			   }
		  }
	   }
			
	Fsetdta(savessp);	/* restore DTA address */
	return(fatal);		/*  Return operation status */


}


/*------------------------------------------------------*/
/*      end program & close window & workstation        */
/*------------------------------------------------------*/
                     
end()

{
	menu_bar(menu_add,0);	/*  turn off menubar */
	appl_exit();		/* So long GEM !  */
}



/**************************  THE END   **********************************/
